import builtins
import re
from typing import Optional, Union

from fastapi import Query

from .type_aliases import JSON, EllipsisType

# NOTE: The regex below is used by fastapi to parse the string representation of a slice
# and raise a 422 error if the string is not valid.
# It does not capture certain erroneous cases, such as ",,", for example.
DIM_REGEX = r"(?:(?:-?\d+)?:){0,2}(?:-?\d+)?"
SLICE_REGEX = rf"^{DIM_REGEX}(?:,{DIM_REGEX})*$"
# SLICE_REGEX = rf"^(?:!,){DIM_REGEX}(?:,(?:=[^,]){DIM_REGEX})*$"


class NDSlice(tuple):
    """A representation of a slice for N-dimensional arrays.

    This class is a tuple of integers, slices, and Ellipsis objects. It is used to
    represent a slice of an N-dimensional array, similar to how numpy uses slices.

    It also provides methods to convert to and from JSON and numpy-style string
    representations of slices. To encode Ellipsis in JSON, it is represented as
    {"step": 0}, which is not a valid builtin.slice.
    """

    def __new__(cls, *args: Union[int, builtins.slice, EllipsisType]) -> "NDSlice":
        if any(not isinstance(s, (int, builtins.slice, EllipsisType)) for s in args):
            raise TypeError(
                f"NDSlice expected int, slice or Ellipsis; got {args} instead. Use "
                "NDSlice.from_numpy_str() or NDSlice.from_json() to parse strings or JSON."
            )
        if len([None for x in args if x == Ellipsis]) > 1:
            raise ValueError(
                f"NDSlice can only contain one Ellipsis; got {args} instead."
            )

        return super().__new__(cls, tuple(args))

    def __bool__(self) -> bool:
        "NDSlice is considered empty if all slices are '...' or ':', i.e. returning entire array"
        full_slices = (
            builtins.slice(None),
            builtins.slice(0, None),
            builtins.slice(0, None, 1),
            builtins.slice(None, None, 1),
            Ellipsis,
        )
        return bool(super()) and not all(
            isinstance(s, EllipsisType) or s in full_slices for s in self
        )

    @classmethod
    def from_json(cls, ser: list[JSON]) -> "NDSlice":
        "Deserialize a json representation of an NDSlice"
        result = []
        for s in ser:
            if isinstance(s, dict):
                result.append(
                    builtins.slice(s.get("start"), s.get("stop"), s.get("step"))
                )
            elif isinstance(s, int):
                result.append(s)
            else:
                raise ValueError("Can not initialize NDSlice from %s", s)
        return cls(*result)

    def to_json(self, ndim: Optional[int] = None) -> list[JSON]:
        "Convert NDSlice into a JSON-serializable representation"
        if not ndim and (Ellipsis in self) and self[-1] != Ellipsis:
            raise ValueError(
                "Converting to JSON an NDSlice with Ellipsis in other than the last element "
                "requires the number of dimensions `ndim` to be specified."
            )
        elif ndim is not None and ndim < len([x for x in self if x != Ellipsis]):
            raise ValueError(
                "Specified number of dimensions `ndim` is less than the number of elements."
            )
        ndim = ndim or len(self)

        def convert_dimension(slc: Union[int, builtins.slice]) -> JSON:
            if isinstance(slc, builtins.slice):
                return (
                    ({} if slc.start is None else {"start": slc.start})
                    | ({} if slc.stop is None else {"stop": slc.stop})
                    | ({} if slc.step is None else {"step": slc.step})
                )
            elif isinstance(slc, int):
                return slc

        fwd, bwd = [], []
        for s in self:
            if s == Ellipsis:
                break
            fwd.append(convert_dimension(s))
        for s in reversed(self[len(fwd) + 1 :]):  # noqa: E203
            bwd.append(convert_dimension(s))

        return fwd + [{} for _ in range(ndim - len(fwd) - len(bwd))] + [*reversed(bwd)]

    @classmethod
    def from_numpy_str(cls, arg: str) -> "NDSlice":
        """Parse and convert a numpy-style string representation of a slice

        For example, '(1:3, 4, 1:5:2, ...)' is converted to (slice(1, 3), 4, slice(1, 5, 2), ...).
        """

        arg = arg.strip("(][)").replace(" ", "")

        if not arg:
            return cls()

        result = []
        for part in arg.split(","):
            if part.strip() == "...":
                result.append(Ellipsis)
            elif part.strip() in {":", "::"}:
                result.append(builtins.slice(None))
            elif m := re.match(r"^(-?\d+)::?$", part.strip()):
                # "start:" or "start::"
                start, stop = int(m.group(1)), None
                result.append(builtins.slice(start, stop))
            elif m := re.match(r"^:(-?\d+):?$", part.strip()):
                # ":stop" or ":stop:"
                start, stop = None, int(m.group(1))
                result.append(builtins.slice(start, stop))
            elif m := re.match(r"^(-?\d+):(-?\d+):(-?\d+)$", part.strip()):
                # "start:stop:step"
                start, stop, step = map(int, m.groups())
                result.append(builtins.slice(start, stop, step))
            elif m := re.match(r"^(-?\d+):(-?\d+):*$", part.strip()):
                # "start:stop" or "start:stop:"
                start, stop = map(int, m.groups())
                result.append(builtins.slice(start, stop))
            elif m := re.match(r"^:(-?\d+):(-?\d+)$", part.strip()):
                # ":stop:step"
                stop, step = map(int, m.groups())
                result.append(builtins.slice(None, stop, step))
            elif m := re.match(r"^::(-?\d+)$", part.strip()):
                # "::step"
                step = int(m.group(1))
                result.append(builtins.slice(None, None, step))
            elif m := re.match(r"^(-?\d+)::(-?\d+)$", part.strip()):
                # "start::step"
                start, step = map(int, m.groups())
                result.append(builtins.slice(start, None, step))
            elif m := re.match(r"^(-?\d+)$", part.strip()):
                result.append(int(m.group()))
            else:
                raise ValueError(f'Invalid slice part: "{part}"')
        return cls(*result)

    @classmethod
    def from_query(cls, slice: str = Query("", pattern=SLICE_REGEX)) -> "NDSlice":
        """Parse and convert a query parameter representation of a slice"""
        return cls.from_numpy_str(slice)

    def to_numpy_str(self) -> str:
        "Convert NDSlice to a numpy-style string representation"
        result = []
        for s in self:
            if isinstance(s, builtins.slice):
                string_slice = (
                    f"{(s.start or '')}:"
                    + ("" if s.stop is None else f"{s.stop}")
                    + (f":{str(s.step)}" if s.step else "")
                )
                result.append(string_slice)
            elif isinstance(s, int):
                result.append(str(s))
            elif s is Ellipsis:
                result.append("...")
            else:
                raise ValueError("Unprocessable entry in NDSlice, %s", s)
        return ",".join(result)



"""
This module handles client configuration.

See config.py for server configuration.

It contains several functions that are factored to facilitate testing,
but the user-facing functionality is straightforward.
"""

import collections
import collections.abc
import os
import shutil
import sys
import warnings
from functools import cache
from pathlib import Path

import jsonschema
import platformdirs

from .utils import parse

TILED_CACHE_DIR = Path(
    os.getenv("TILED_CACHE_DIR", platformdirs.user_cache_dir("tiled"))
)
__all__ = [
    "list_profiles",
    "load_profiles",
    "paths",
    "create_profile",
    "delete_profile",
    "set_default_profile_name",
    "get_default_profile_name",
]


@cache
def schema():
    "Load the schema for profiles."
    import yaml

    here = Path(__file__).parent.absolute()
    schema_path = here / "config_schemas" / "client_profiles.yml"
    with open(schema_path, "r") as file:
        return yaml.safe_load(file)


# Some items in the search path is system-dependent, and others are hard-coded.
# Paths later in the list ("closer" to the user) have higher precedence.
_all_paths = [
    Path(
        os.getenv("TILED_SITE_PROFILES", Path("/etc/tiled/profiles"))
    ),  # hard-coded system path
    Path(
        os.getenv(
            "TILED_SITE_PROFILES",
            Path(platformdirs.site_config_dir("tiled"), "profiles"),
        )
    ),  # XDG-compliant system path
    Path(sys.prefix, "etc", "tiled", "profiles"),  # environment
    Path(
        os.getenv("TILED_PROFILES", Path.home() / ".config/tiled/profiles")
    ),  # hard-coded user path
    Path(
        os.getenv(
            "TILED_PROFILES", Path(platformdirs.user_config_dir("tiled"), "profiles")
        )
    ),  # system-dependent user path
]
# Remove duplicates (i.e. if XDG and hard-coded are the same on this system).
_seen = set()
paths = [x for x in _all_paths if not (x in _seen or _seen.add(x))]
del _seen


def gather_profiles(paths, strict=True):
    """
    For each path in paths, return a dict mapping filepath to content.
    """
    levels = []
    for path in paths:
        filepath_to_content = {}
        if path.is_dir():
            for filename in path.iterdir():
                filepath = path / filename
                # Ignore hidden files and .py files.
                if (
                    filename.name.startswith(".")
                    or filename.suffix == ".py"
                    or filename.name == "__pycache__"
                ):
                    continue
                try:
                    with open(filepath) as file:
                        content = parse(file)
                        if content is None:
                            raise ProfileError(f"File {filepath!s} is empty.")
                        if not isinstance(content, collections.abc.Mapping):
                            raise ProfileError(
                                f"File {filepath!s} does not have the expected structure."
                            )
                        for profile_name, profile_content in content.items():
                            try:
                                jsonschema.validate(
                                    instance=profile_content, schema=schema()
                                )
                            except jsonschema.ValidationError as validation_err:
                                original_msg = validation_err.args[0]
                                raise ProfileError(
                                    f"ValidationError while parsing profile {profile_name} "
                                    f"in file {filepath!s}: {original_msg}"
                                ) from validation_err
                except Exception as err:
                    if strict:
                        raise
                    else:
                        warnings.warn(
                            f"Skipping {filepath!s}. Failed to parse with error: {err!r}."
                        )
                        continue
                if not isinstance(content, collections.abc.Mapping):
                    if strict:
                        raise
                    else:
                        warnings.warn(
                            f"Skipping {filepath!r}. Content has invalid structure (not a mapping)."
                        )
                filepath_to_content[filepath] = content
        levels.append(filepath_to_content)
    return levels


def resolve_precedence(levels):
    """
    Given a list of mappings (filename-to-content), resolve precedence.

    In the event of irresolvable collisions, drop the offenders and warn.

    The result is a mapping from profile name to (filepath, content).
    """
    # Map each profile_name to the file(s) that define a profile with that name.
    # This is used to track collisions.
    profile_name_to_filepaths_per_level = [
        collections.defaultdict(list) for _ in range(len(paths))
    ]
    for profile_name_to_filepaths, filepath_to_content in zip(
        profile_name_to_filepaths_per_level, levels
    ):
        for filepath, content in filepath_to_content.items():
            for profile_name in content:
                profile_name_to_filepaths[profile_name].append(filepath)
    combined = {}
    collisions = {}
    for profile_name_to_filepaths, filepath_to_content in zip(
        profile_name_to_filepaths_per_level, levels
    ):
        for profile_name, filepaths in profile_name_to_filepaths.items():
            # A profile name in this level resolves any collisions in the previous level.
            collisions.pop(profile_name, None)
            # Does more than one file *in this same level (directory) in the search path*
            # define a profile with the same name? If so, there is no sure way to decide
            # precedence so we will omit it and issue a warning, unless something later
            # in the search path overrides it.
            if len(filepaths) > 1:
                # Stash collision for possible warning below.
                collisions[profile_name] = filepaths
                # If a previous level defined this, remove it.
                combined.pop(profile_name, None)
            else:
                (filepath,) = filepaths
                combined[profile_name] = (
                    filepath,
                    filepath_to_content[filepath][profile_name],
                )
    MSG = """More than file in the same directory:

{filepaths}

defines a profile with the name {profile_name!r}.

The profile will be ommitted. Fix this by removing one of the duplicates"""
    for profile_name, filepaths in collisions.items():
        # Is this file in either the XDG user or hard-coded user directory
        # (which might or might not be the same directory)?
        if filepaths[0].is_relative_to(_all_paths[-1]) or filepaths[0].is_relative_to(
            _all_paths[-2]
        ):
            msg = (MSG + ".").format(
                filepaths="\n".join(map(str, filepaths)), profile_name=profile_name
            )
        else:
            msg = (
                MSG
                + (
                    " or by defining a profile with that name in a "
                    f"file in the user config directory {paths[-1]} "
                    "to override them."
                )
            ).format(
                filepaths="\n".join(map(str, filepaths)), profile_name=profile_name
            )
        warnings.warn(msg)
    return combined


@cache
def load_profiles():
    """
    Return a mapping of profile_name to (source_path, content).

    The files are only actually read the first time this is called.
    Thereafter, the results are cached. To clear the cache and re-read,
    use load_profiles.cache_clear().

    The search path for the source files is available from Python as:

    >>> tiled.client.profiles.paths

    or from a CLI as:

    $ tiled profile paths

    See also

    $ tiled profile list
    $ tiled profile show PROFILE_NAME
    """
    levels = gather_profiles(paths, strict=False)
    profiles = resolve_precedence(levels)
    return profiles


def list_profiles():
    """
    Return a mapping of profile names to source filepath.

    The search path for the source files is available from Python as:

    >>> tiled.client.profiles.paths

    or from a CLI as:

    $ tiled profile paths

    See also

    $ tiled profile list
    $ tiled profile show PROFILE_NAME
    """
    levels = gather_profiles(paths, strict=False)
    profiles = resolve_precedence(levels)
    return {name: source_filepath for name, (source_filepath, _) in profiles.items()}


def _compose_profile(name, *, uri, verify):
    "Compose profile YAML."
    import yaml

    content = {name: {"uri": uri, "verify": verify}}
    return yaml.dump(content)


def create_profile(uri, name, verify=True, overwrite=False):
    """
    Create a new profile.

    This only includes the most commonly-used options. (More could be added.)
    """
    text = _compose_profile(name, uri=uri, verify=verify)
    filepath = paths[-1] / f"{name}.yml"
    filepath.parent.mkdir(parents=True, exist_ok=True)
    if overwrite:
        mode = "wt"
    else:
        mode = "xt"
    try:
        with open(filepath, mode) as file:
            file.write(text)
    except FileExistsError:
        raise ProfileExists(
            f"Profile named {name} already exists at {filepath}. "
            "Use overwite=True to overwrite it."
        )
    return filepath


def delete_profile(name):
    """
    Delete a profile by name.

    This will walk the search path, starting with the highest precedence
    directory, and delete only the first match it finds.
    """
    for path in paths:
        # All profiles created by create_profile have extension .yml but
        # a user-written one may have extension .yaml.
        for ext in {".yml", ".yaml"}:
            filepath = path / f"{name}{ext}"
            if filepath.exists():
                filepath.unlink()
                return filepath


def get_default_profile_name():
    """
    Return the name of the current default profile.
    """
    filepath = paths[-1].parent / "default_profile"
    try:
        return filepath.read_text()
    except FileNotFoundError:
        return None


def set_default_profile_name(name):
    filepath = paths[-1].parent / "default_profile"
    filepath.parent.mkdir(parents=True, exist_ok=True)
    if name is None:
        if filepath.exists():
            filepath.unlink()
        return
    if name not in list_profiles():
        raise ProfileNotFound(name)
    with open(filepath, "w") as file:
        file.write(name)
    # Clean up cruft from older versions of Tiled.
    UNUSED_DIR = TILED_CACHE_DIR / "default_identities"
    if UNUSED_DIR.is_dir():
        shutil.rmtree(UNUSED_DIR)


class ProfileNotFound(KeyError):
    pass


class ProfileError(ValueError):
    pass


class ProfileExists(ValueError):
    pass



"""Add access_tags to API keys

Revision ID: a806cc635ab2
Revises: 0c705a02954c
Create Date: 2025-08-26 17:10:47.717942

"""
import sqlalchemy as sa
from alembic import op

from tiled.authn_database.orm import APIKey, JSONList

# revision identifiers, used by Alembic.
revision = "a806cc635ab2"
down_revision = "0c705a02954c"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        APIKey.__tablename__,
        sa.Column("access_tags", JSONList(511), nullable=True),
    )


def downgrade():
    op.drop_column(
        APIKey.__tablename__,
        "access_tags",
    )



import sys
import threading

import numpy as np
import pytest
from starlette.testclient import WebSocketDenialResponse

from ..catalog import from_uri
from ..client import Context, from_context
from ..client.stream import Subscription
from ..server.app import build_app

pytestmark = pytest.mark.skipif(
    sys.platform == "win32", reason="Requires Redis service"
)


@pytest.fixture(scope="function")
def tiled_websocket_context(tmpdir, redis_uri):
    """Fixture that provides a Tiled context with websocket support."""

    tree = from_uri(
        "sqlite:///:memory:",
        writable_storage=[
            f"file://localhost{str(tmpdir / 'data')}",
            f"duckdb:///{tmpdir / 'data.duckdb'}",
        ],
        readable_storage=None,
        init_if_not_exists=True,
        cache_settings={
            "uri": redis_uri,
            "ttl": 60,
            "socket_timeout": 10.0,
            "socket_connect_timeout": 10.0,
        },
    )

    app = build_app(
        tree,
        authentication={"single_user_api_key": "secret"},
    )

    with Context.from_app(app) as context:
        yield context


def test_subscribe_immediately_after_creation_websockets(tiled_websocket_context):
    context = tiled_websocket_context
    client = from_context(context)

    # Create streaming array node using Tiled client
    arr = np.arange(10)
    streaming_node = client.write_array(arr, key="test_stream_immediate")

    # Set up subscription using the Subscription class
    received = []
    received_event = threading.Event()

    def callback(subscription, data):
        """Callback to collect received messages."""
        received.append(data)
        if len(received) >= 3:
            received_event.set()

    # Create subscription for the streaming node
    subscription = Subscription(
        context=context,
        segments=["test_stream_immediate"],
    )
    subscription.add_callback(callback)

    # Start the subscription
    subscription.start()

    # Write updates using Tiled client
    for i in range(1, 4):
        new_arr = np.arange(10) + i
        streaming_node.write(new_arr)

    # Wait for all messages to be received
    assert received_event.wait(timeout=5.0), "Timeout waiting for messages"

    # Verify all updates received in order
    assert len(received) == 3

    # Check that we received messages with the expected data
    for i, msg in enumerate(received):
        assert "timestamp" in msg
        assert "payload" in msg
        assert msg["shape"] == [10]

        # Verify payload contains the expected array data
        payload_array = np.frombuffer(msg["payload"], dtype=np.int64)
        expected_array = np.arange(10) + (i + 1)
        np.testing.assert_array_equal(payload_array, expected_array)

    # Clean up the subscription
    subscription.stop()


def test_websocket_connection_to_non_existent_node_subscription(
    tiled_websocket_context,
):
    """Test subscription to non-existent node raises appropriate error."""
    context = tiled_websocket_context

    non_existent_node_id = "definitely_non_existent_websocket_node_99999999"

    # Create subscription for non-existent node
    subscription = Subscription(
        context=context,
        segments=[non_existent_node_id],
    )

    # Attempting to start should raise WebSocketDenialResponse
    with pytest.raises(WebSocketDenialResponse):
        subscription.start()


def test_subscribe_after_first_update_subscription(tiled_websocket_context):
    """Client that subscribes after first update sees only subsequent updates."""
    context = tiled_websocket_context
    client = from_context(context)

    # Create streaming array node using Tiled client
    arr = np.arange(10)
    streaming_node = client.write_array(arr, key="test_stream_after_update")

    # Write first update before subscribing
    first_update = np.arange(10) + 1
    streaming_node.write(first_update)

    # Set up subscription using the Subscription class (no start parameter = only new updates)
    received = []
    received_event = threading.Event()

    def callback(subscription, data):
        """Callback to collect received messages."""
        received.append(data)
        if len(received) >= 2:
            received_event.set()

    # Create subscription for the streaming node
    subscription = Subscription(
        context=context,
        segments=["test_stream_after_update"],
    )
    subscription.add_callback(callback)

    # Start the subscription
    subscription.start()

    # Write more updates
    for i in range(2, 4):
        new_arr = np.arange(10) + i
        streaming_node.write(new_arr)

    # Wait for messages to be received
    assert received_event.wait(timeout=5.0), "Timeout waiting for messages"

    # Should only receive the 2 new updates (not the first one)
    assert len(received) == 2

    # Check that we received messages with the expected data
    for i, msg in enumerate(received):
        assert "timestamp" in msg
        assert "payload" in msg
        assert msg["shape"] == [10]

        # Verify payload contains the expected array data
        payload_array = np.frombuffer(msg["payload"], dtype=np.int64)
        expected_array = np.arange(10) + (i + 2)  # i+2 because we start from update 2
        np.testing.assert_array_equal(payload_array, expected_array)

    # Clean up the subscription
    subscription.stop()


def test_subscribe_after_first_update_from_beginning_subscription(
    tiled_websocket_context,
):
    """Client that subscribes after first update but requests from start=0 sees all updates."""
    context = tiled_websocket_context
    client = from_context(context)

    # Use unique key to avoid interference from other tests
    import uuid

    unique_key = f"test_stream_from_beginning_{uuid.uuid4().hex[:8]}"

    # Create streaming array node using Tiled client
    arr = np.arange(10)
    streaming_node = client.write_array(arr, key=unique_key)

    # Write first update before subscribing
    first_update = np.arange(10) + 1
    streaming_node.write(first_update)

    # Set up subscription using the Subscription class with start=0
    received = []
    received_event = threading.Event()

    def callback(subscription, data):
        """Callback to collect received messages."""
        received.append(data)
        if len(received) >= 4:  # initial + first update + 2 new updates
            received_event.set()

    # Create subscription for the streaming node with start=0
    subscription = Subscription(context=context, segments=[unique_key], start=0)
    subscription.add_callback(callback)

    # Start the subscription
    subscription.start()

    # Write more updates
    for i in range(2, 4):
        new_arr = np.arange(10) + i
        streaming_node.write(new_arr)

    # Wait for all messages to be received
    assert received_event.wait(timeout=5.0), "Timeout waiting for messages"

    # Should receive: initial array + first update + 2 new updates = 4 total
    assert len(received) == 4

    # Check the messages in order
    # First message: initial array creation
    msg = received[0]
    assert "timestamp" in msg
    assert "payload" in msg
    assert msg["shape"] == [10]
    payload_array = np.frombuffer(msg["payload"], dtype=np.int64)
    expected_array = np.arange(10)  # Initial array
    np.testing.assert_array_equal(payload_array, expected_array)

    # Remaining messages: updates 1, 2, 3
    for i, msg in enumerate(received[1:], 1):
        assert "timestamp" in msg
        assert "payload" in msg
        assert msg["shape"] == [10]

        # Verify payload contains the expected array data
        payload_array = np.frombuffer(msg["payload"], dtype=np.int64)
        expected_array = np.arange(10) + i
        np.testing.assert_array_equal(payload_array, expected_array)

    # Clean up the subscription
    subscription.stop()



import tempfile

import pyarrow as pa
import pytest

from tiled.adapters.arrow import ArrowAdapter
from tiled.storage import FileStorage
from tiled.structures.core import StructureFamily
from tiled.structures.data_source import DataSource, Management
from tiled.structures.table import TableStructure

names = ["f0", "f1", "f2"]
data0 = [
    pa.array([1, 2, 3, 4, 5]),
    pa.array(["foo0", "bar0", "baz0", None, "goo0"]),
    pa.array([True, None, False, True, None]),
]
data1 = [
    pa.array([6, 7, 8, 9, 10, 11, 12]),
    pa.array(["foo1", "bar1", None, "baz1", "biz", None, "goo"]),
    pa.array([None, True, True, False, False, None, True]),
]
data2 = [pa.array([13, 14]), pa.array(["foo2", "baz2"]), pa.array([False, None])]

batch0 = pa.record_batch(data0, names=names)
batch1 = pa.record_batch(data1, names=names)
batch2 = pa.record_batch(data2, names=names)
data_uri = "file://localhost/" + tempfile.gettempdir()


@pytest.fixture
def data_source_from_init_storage() -> DataSource[TableStructure]:
    table = pa.Table.from_arrays(data0, names)
    structure = TableStructure.from_arrow_table(table, npartitions=3)
    data_source = DataSource(
        management=Management.writable,
        mimetype="application/vnd.apache.arrow.file",
        structure_family=StructureFamily.table,
        structure=structure,
        assets=[],
    )
    storage = FileStorage(data_uri)
    return ArrowAdapter.init_storage(
        data_source=data_source, storage=storage, path_parts=[]
    )


@pytest.fixture
def adapter(data_source_from_init_storage: DataSource[TableStructure]) -> ArrowAdapter:
    data_source = data_source_from_init_storage
    return ArrowAdapter(
        [asset.data_uri for asset in data_source.assets],
        data_source.structure,
    )


def test_attributes(adapter: ArrowAdapter) -> None:
    assert adapter.structure().columns == names
    assert adapter.structure().npartitions == 3


def test_write_read(adapter: ArrowAdapter) -> None:
    # test writing to a partition and reading it
    adapter.write_partition(batch0, 0)
    assert pa.Table.from_arrays(data0, names) == pa.Table.from_pandas(
        adapter.read_partition(0)
    )

    adapter.write_partition([batch0, batch1], 1)
    assert pa.Table.from_batches([batch0, batch1]) == pa.Table.from_pandas(
        adapter.read_partition(1)
    )

    adapter.write_partition([batch0, batch1, batch2], 2)
    assert pa.Table.from_batches([batch0, batch1, batch2]) == pa.Table.from_pandas(
        adapter.read_partition(2)
    )

    # test write to all partitions and read all
    adapter.write_partition([batch0, batch1, batch2], 0)
    adapter.write_partition([batch2, batch0, batch1], 1)
    adapter.write_partition([batch1, batch2, batch0], 2)

    assert pa.Table.from_pandas(adapter.read()) == pa.Table.from_batches(
        [batch0, batch1, batch2, batch2, batch0, batch1, batch1, batch2, batch0]
    )

    # test adapter.write() raises NotImplementedError when there are more than 1 partitions
    with pytest.raises(NotImplementedError):
        adapter.write(batch0)



import awkward

from ..serialization.awkward import from_zipped_buffers, to_zipped_buffers
from ..structures.awkward import project_form
from .base import BaseClient
from .utils import export_util, handle_error, retry_context


class AwkwardClient(BaseClient):
    def __repr__(self):
        # TODO Include some summary of the structure. Probably
        # lift __repr__ code from awkward itself here.
        form = awkward.forms.from_dict(self.structure().form)
        if isinstance(form, awkward.forms.recordform.RecordForm):
            out = f"<{type(self).__name__} {{"
            # Always show at least one field
            if form.fields:
                out += form.fields[0]
            # And then show as many more as we can fit on one line.
            counter = 1
            for sample_repr in form.fields[1:]:
                if len(out) + len(sample_repr) > 60:  # character count
                    break
                out += ", " + sample_repr
                counter += 1
            # Are there more fields that what we displayed above?
            if len(form.fields) > counter:
                out += f", ...}} ~{len(form.fields)} field"
                out += "s," if len(form.fields) > 1 else ","
            else:
                out += "}"
            out += f" {self.structure().length} item"
            out += "s>" if self.structure().length > 1 else ">"
            return out

        return f"<{type(self).__name__}>"

    def write(self, container):
        structure = self.structure()
        components = (structure.form, structure.length, container)
        for attempt in retry_context():
            with attempt:
                handle_error(
                    self.context.http_client.put(
                        self.item["links"]["full"],
                        content=bytes(
                            to_zipped_buffers("application/zip", components, {})
                        ),
                        headers={"Content-Type": "application/zip"},
                    )
                )

    def read(self, slice=...):
        structure = self.structure()
        form = awkward.forms.from_dict(structure.form)
        typetracer, report = awkward.typetracer.typetracer_with_report(form)
        proxy_array = awkward.Array(typetracer)
        awkward.typetracer.touch_data(proxy_array[slice])
        form_keys_touched = set(report.data_touched)
        projected_form = project_form(form, form_keys_touched)
        # The order is not important, but sort so that the request is deterministic.
        form_keys = sorted(form_keys_touched)
        for attempt in retry_context():
            with attempt:
                content = handle_error(
                    self.context.http_client.post(
                        self.item["links"]["buffers"],
                        headers={"Accept": "application/zip"},
                        json=form_keys,
                    )
                ).read()
        container = from_zipped_buffers(content, projected_form, structure.length)
        projected_array = awkward.from_buffers(
            projected_form, structure.length, container, allow_noncanonical_form=True
        )
        return projected_array[slice]

    def __getitem__(self, slice):
        return self.read(slice=slice)

    def export(self, filepath, *, format=None):
        """
        Download data in some format and write to a file.

        Parameters
        ----------
        file: str or buffer
            Filepath or writeable buffer.
        format : str, optional
            If format is None and `file` is a filepath, the format is inferred
            from the name, like 'table.csv' implies format="text/csv". The format
            may be given as a file extension ("csv") or a media type ("text/csv").

        Examples
        --------

        Export as Parquet.

        >>> a.export("awkward.parquet")

        Export as JSON.

        >>> a.export("awkward.json")

        """
        return export_util(
            filepath,
            format,
            self.context.http_client.get,
            self.item["links"]["full"],
            params={},
        )



import numpy

from tiled.client import Context, from_context
from tiled.server.app import build_app_from_config


def test_mount_node(sqlite_or_postgres_uri, tmpdir):
    "Test 'mounting' sub-trees of a catalog."
    one_tree_config = {
        "trees": [
            {
                "path": "/",
                "tree": "catalog",
                "args": {
                    "uri": sqlite_or_postgres_uri,
                    "init_if_not_exists": True,
                    "writable_storage": [tmpdir / "data"],
                },
            },
        ]
    }
    # Make two containers in the catalog database, A and B.
    with Context.from_app(build_app_from_config(one_tree_config)) as context:
        client = from_context(context)
        client.create_container("A")
        client.create_container("B")

    # 'Mount' the two container nodes A and B at /a and /b respectively.
    # Now it is as if we have two separate catalogs, but they happen
    # to be housed in one database.
    multi_tree_config = {
        "trees": [
            {
                "path": "/a",
                "tree": "catalog",
                "args": {
                    "uri": sqlite_or_postgres_uri,
                    "writable_storage": [tmpdir / "data"],
                    "mount_node": "/A",
                },
            },
            {
                "path": "/b",
                "tree": "catalog",
                "args": {
                    "uri": sqlite_or_postgres_uri,
                    "writable_storage": [tmpdir / "data"],
                    "mount_node": "/B",
                },
            },
        ]
    }
    with Context.from_app(build_app_from_config(multi_tree_config)) as context:
        client = from_context(context)
        # Create a new node in each section of the catalog.
        client["a"].create_container("x")
        client["a"]["x"].create_container("i")
        client["b"].create_container("y")
        # Check for no cross-talk.
        assert list(client) == ["a", "b"]
        assert list(client["a"]) == ["x"]
        assert list(client["a"]["x"]) == ["i"]
        assert list(client["b"]) == ["y"]
        # Check that assets do not collide.
        arr1 = numpy.array([1, 2, 3])
        arr2 = numpy.array([4, 5, 6])
        ac1 = client["a"].write_array(arr1, key="c")
        ac2 = client["b"].write_array(arr2, key="c")
        assert numpy.array_equal(ac1[:], arr1)
        assert numpy.array_equal(ac2[:], arr2)
        uri1 = ac1.data_sources()[0].assets[0].data_uri
        uri2 = ac2.data_sources()[0].assets[0].data_uri
        assert uri1 != uri2

    # Mount a more deeply nested node at a nested path.
    multi_tree_config = {
        "trees": [
            {
                "path": "/some/nested/path",
                "tree": "catalog",
                "args": {
                    "uri": sqlite_or_postgres_uri,
                    "writable_storage": [tmpdir / "data"],
                    "mount_node": "/A/x",
                },
            },
        ]
    }
    with Context.from_app(build_app_from_config(multi_tree_config)) as context:
        client = from_context(context)
        assert list(client["some"]["nested"]["path"]) == ["i"]

    # As above, but specify the mount_node as a list of path segments.
    multi_tree_config = {
        "trees": [
            {
                "path": "/some/nested/path",
                "tree": "catalog",
                "args": {
                    "uri": sqlite_or_postgres_uri,
                    "writable_storage": [tmpdir / "data"],
                    "mount_node": ["A", "x"],
                },
            },
        ]
    }
    with Context.from_app(build_app_from_config(multi_tree_config)) as context:
        client = from_context(context)
        assert list(client["some"]["nested"]["path"]) == ["i"]



import contextlib
import os.path
import tempfile
from collections.abc import Generator
from pathlib import Path
from typing import Iterable, Optional

from alembic import command
from alembic.config import Config
from alembic.runtime import migration
from sqlalchemy.ext.asyncio import AsyncEngine


def write_alembic_ini(
    alembic_ini_template_path: Path, alembic_dir: Path, path: Path, database_uri: str
):
    """Write a complete alembic.ini from our template.

    Parameters
    ----------
    alembic_ini_template_path: str
    alembic_dir: str
    path : str
        path to the alembic.ini file that should be written.
    db_url : str
        The SQLAlchemy database url, e.g. `sqlite+aiosqlite:///tiled.sqlite`.
    """
    with open(alembic_ini_template_path) as f:
        alembic_ini_tpl = f.read()

    with open(path, "w") as f:
        f.write(
            alembic_ini_tpl.format(
                migration_script_directory=alembic_dir,
                # If there are any %s in the URL, they should be replaced with %%, since ConfigParser
                # by default uses %() for substitution. You'll get %s in your URL when you have usernames
                # with special chars (such as '@') that need to be URL encoded. URL Encoding is done with %s.
                # YAY for nested templates?
                database_uri=str(database_uri).replace("%", "%%"),
            )
        )


@contextlib.contextmanager
def temp_alembic_ini(
    alembic_ini_template_path: Path, alembic_dir: Path, database_uri: str
) -> Generator[str]:
    """
    Context manager for temporary alembic configuration file

    Temporarily write an alembic.ini file for use with alembic migration scripts.
    Context manager yields alembic.ini path.

    Parameters
    ----------
    alembic_ini_template_path: str
    alembic_dir: str
    database_uri : str
        The SQLAlchemy database url, e.g. `sqlite+aiosqlite:///tiled.sqlite`.

    Returns
    -------
    alembic_ini: str
        The path to the temporary alembic.ini that we have created.
        This file will be cleaned up on exit from the context manager.
    """
    with tempfile.TemporaryDirectory() as td:
        alembic_ini = os.path.join(td, "alembic.ini")
        write_alembic_ini(
            alembic_ini_template_path, alembic_dir, alembic_ini, database_uri
        )
        yield alembic_ini


def stamp_head(alembic_ini_template_path: Path, alembic_dir: Path, engine_url: str):
    """
    Upgrade schema to the specified revision.
    """
    with temp_alembic_ini(
        alembic_ini_template_path, alembic_dir, engine_url
    ) as alembic_ini:
        alembic_cfg = Config(alembic_ini)
        command.stamp(alembic_cfg, "head")


def upgrade(
    alembic_ini_template_path: Path, alembic_dir: Path, engine_url: str, revision: str
):
    """
    Upgrade schema to the specified revision.
    """
    with temp_alembic_ini(
        alembic_ini_template_path, alembic_dir, engine_url
    ) as alembic_ini:
        alembic_cfg = Config(alembic_ini)
        command.upgrade(alembic_cfg, revision)


def downgrade(
    alembic_ini_template_path: Path, alembic_dir: Path, engine_url: str, revision: str
):
    """
    Downgrade schema to the specified revision.
    """
    with temp_alembic_ini(
        alembic_ini_template_path, alembic_dir, engine_url
    ) as alembic_ini:
        alembic_cfg = Config(alembic_ini)
        command.downgrade(alembic_cfg, revision)


class UnrecognizedDatabase(Exception):
    pass


class UninitializedDatabase(Exception):
    pass


class DatabaseUpgradeNeeded(Exception):
    pass


async def get_current_revision(
    engine: AsyncEngine, known_revisions: Iterable[str]
) -> Optional[str]:
    redacted_url = engine.url._replace(password="[redacted]")
    async with engine.connect() as conn:
        context = await conn.run_sync(migration.MigrationContext.configure)
        heads = await conn.run_sync(lambda conn: context.get_current_heads())
    if heads == ():
        return None
    elif len(heads) != 1:
        raise UnrecognizedDatabase(
            f"This database {redacted_url} is stamped with an alembic revisions {heads}. "
            "It looks like Tiled has been configured to connect to a database "
            "already populated by some other application (not Tiled) or else "
            "its database is in a corrupted state."
        )
    (revision,) = heads
    if revision not in known_revisions:
        raise UnrecognizedDatabase(
            f"The database {redacted_url} has an unrecognized revision {revision}. "
            "It may have been created by a newer version of Tiled."
        )
    return revision


async def check_database(
    engine: AsyncEngine, required_revision: str, known_revisions: Iterable[str]
):
    revision = await get_current_revision(engine, known_revisions)
    redacted_url = engine.url._replace(password="[redacted]")
    if revision is None:
        raise UninitializedDatabase(
            f"The database {redacted_url} has no revision stamp. It may be empty. "
            "It can be initialized with `initialize_database(engine)`."
        )
    elif revision != required_revision:
        raise DatabaseUpgradeNeeded(
            f"The database {redacted_url} has revision {revision} and "
            f"needs to be upgraded to revision {required_revision}."
        )



import builtins
from typing import Any, Dict, List, Optional, Set, Tuple, Union, cast

import tifffile
from numpy._typing import NDArray

from ..catalog.orm import Node
from ..ndslice import NDSlice
from ..storage import Storage
from ..structures.array import ArrayStructure, BuiltinDtype
from ..structures.core import Spec, StructureFamily
from ..structures.data_source import DataSource
from ..type_aliases import JSON
from ..utils import path_from_uri
from .resource_cache import with_resource_cache
from .sequence import FileSequenceAdapter
from .utils import init_adapter_from_catalog


class TiffAdapter:
    """
    Read a TIFF file.

    Examples
    --------

    >>> TiffAdapter("path/to/file.tiff")
    """

    structure_family = StructureFamily.array
    supported_storage: Set[type[Storage]] = set()

    def __init__(
        self,
        data_uri: str,
        *,
        structure: Optional[ArrayStructure] = None,
        metadata: Optional[JSON] = None,
        specs: Optional[List[Spec]] = None,
    ) -> None:
        """

        Parameters
        ----------
        data_uri :
        structure :
        metadata :
        specs :
        """
        filepath = path_from_uri(data_uri)
        cache_key = (tifffile.TiffFile, filepath)
        self._file = with_resource_cache(cache_key, tifffile.TiffFile, filepath)
        self.specs = specs or []
        self._provided_metadata = metadata or {}
        if structure is None:
            if self._file.is_shaped:
                from_file: Tuple[Dict[str, Any], ...] = cast(
                    Tuple[Dict[str, Any], ...], self._file.shaped_metadata
                )
                shape = tuple(from_file[0]["shape"])
            else:
                arr = self._file.asarray()
                shape = arr.shape
            structure = ArrayStructure(
                shape=shape,
                chunks=tuple((dim,) for dim in shape),
                data_type=BuiltinDtype.from_numpy_dtype(self._file.series[0].dtype),
            )
        self._structure = structure

    @classmethod
    def from_catalog(
        cls,
        data_source: DataSource[ArrayStructure],
        node: Node,
        /,
        **kwargs: Optional[Any],
    ) -> "TiffAdapter":
        return init_adapter_from_catalog(cls, data_source, node, **kwargs)  # type: ignore

    @classmethod
    def from_uris(cls, data_uri: str, **kwargs: Optional[Any]) -> "TiffAdapter":
        return cls(data_uri)

    def metadata(self) -> JSON:
        """

        Returns
        -------

        """
        # This contains some enums, but Python's built-in JSON serializer
        # handles them fine (converting  to str or int as appropriate).
        d = {tag.name: tag.value for tag in self._file.pages[0].tags.values()}
        d.update(self._provided_metadata)
        return d

    def read(self, slice: NDSlice = NDSlice(...)) -> NDArray[Any]:
        # TODO Is there support for reading less than the whole array
        # if we only want a slice? I do not think that is possible with a
        # single-page TIFF but I'm not sure. Certainly it *is* possible for
        # multi-page TIFFs.
        arr = self._file.asarray()
        return arr[slice] if slice else arr

    def read_block(
        self, block: Tuple[int, ...], slice: Optional[slice] = None
    ) -> NDArray[Any]:
        # For simplicity, this adapter always treat a single TIFF file as one
        # chunk. This could be relaxed in the future.
        if sum(block) != 0:
            raise IndexError(block)

        arr = self._file.asarray()
        if slice is not None:
            arr = arr[slice]
        return arr

    def structure(self) -> ArrayStructure:
        return self._structure


class TiffSequenceAdapter(FileSequenceAdapter):
    def _load_from_files(
        self, slice: Union[builtins.slice, int] = slice(None)
    ) -> NDArray[Any]:
        return tifffile.TiffSequence(self.filepaths[slice]).asarray()



import copy
import os
import sys
import warnings
from collections.abc import Mapping
from pathlib import Path
from typing import Any, Iterator, List, Optional, Tuple, Union

import dask
import dask.array
import dask.delayed
import h5py
import hdf5plugin  # noqa: F401
import numpy
from numpy._typing import NDArray

from ..adapters.utils import IndexersMixin
from ..catalog.orm import Node
from ..iterviews import ItemsView, KeysView, ValuesView
from ..ndslice import NDSlice
from ..server.core import NoEntry
from ..structures.array import ArrayStructure
from ..structures.core import Spec, StructureFamily
from ..structures.data_source import DataSource
from ..type_aliases import JSON
from ..utils import BrokenLink, Sentinel, node_repr, path_from_uri
from .array import ArrayAdapter

SWMR_DEFAULT = bool(int(os.getenv("TILED_HDF5_SWMR_DEFAULT", "0")))
INLINED_DEPTH = int(os.getenv("TILED_HDF5_INLINED_CONTENTS_MAX_DEPTH", "7"))

HDF5_DATASET = Sentinel("HDF5_DATASET")
HDF5_BROKEN_LINK = Sentinel("HDF5_BROKEN_LINK")


def parse_hdf5_tree(
    tree: Union[h5py.File, h5py.Group, h5py.Dataset]
) -> Union[dict[str, Union[Any, Sentinel]], Sentinel]:
    """Parse an HDF5 file or group into a nested dictionary structure

    the resulting tree structure represents any groups as nested dictionaries ans datasets as None.

    Parameters
    ----------
    tree : h5py.File or h5py.Group
        The file or group to parse

    Returns
    -------
    dict
        A nested dictionary structure representing the HDF5 file or group
    """
    res: dict[str, Union[Any, None]] = {}

    if isinstance(tree, h5py.Dataset):
        return HDF5_DATASET

    for key, val in tree.items():
        res[key] = HDF5_BROKEN_LINK if val is None else parse_hdf5_tree(val)

    return res


def get_hdf5_attrs(
    file_uri: str,
    dataset: Optional[str] = None,
    swmr: bool = SWMR_DEFAULT,
    libver: str = "latest",
    **kwargs: Optional[Any],
) -> JSON:
    """Get attributes of an HDF5 dataset"""
    file_path = path_from_uri(file_uri)
    with h5open(file_path, dataset=dataset, swmr=swmr, libver=libver, **kwargs) as node:
        d = dict(getattr(node, "attrs", {}))
        for k, v in d.items():
            # Convert any bytes to str.
            if isinstance(v, bytes):
                d[k] = v.decode()
    return d


class h5open(h5py.File):  # type: ignore
    """A context manager for reading datasets from HDF5 files

    This class is a subclass of h5py.File that allows for reading datasets from HDF5 files using a context manager.
    It raises a BrokenLink exception if a key referencing a dataset (or a group) exists in the file, but the
    referenced object can not be accessed (e.g. if an externally linked file has been removed). In these cases,
    h5py raises a KeyError with following messages:
    KeyError: 'Unable to synchronously open object (component not found)'
    or
    KeyError: "Unable to synchronously open object (unable to open external file, external link file name = '...')"
    KeyError: "Unable to synchronously open object (can't open file)"
    if a soft link or an external link is broken, respectively.

    This message is distinct from the case when a key does not exist in the file, in which case h5py raises:
    KeyError: "Unable to synchronously open object (object 'y' doesn't exist)"
    """

    def __init__(
        self, filename: Union[str, Path], dataset: Optional[str] = None, **kwargs: Any
    ) -> None:
        super().__init__(filename, mode="r", **kwargs)
        self.dataset = dataset

    def __enter__(self) -> Union[h5py.File, h5py.Group, h5py.Dataset]:
        super().__enter__()
        try:
            return self[self.dataset] if self.dataset else self
        except Exception:
            self.__exit__(*sys.exc_info())
            raise

    def __exit__(self, exc_type, exc_value, exc_tb) -> None:  # type: ignore
        super().__exit__(exc_type, exc_value, exc_tb)

        if exc_type == KeyError:
            if "file" in str(exc_value):
                # External link is broken
                raise BrokenLink(exc_value.args[0]) from exc_value

            elif "component not found" in str(exc_value):
                # Soft link is broken
                raise BrokenLink(exc_value.args[0]) from exc_value


class HDF5ArrayAdapter(ArrayAdapter):
    """Adapter for array-type data stored in HDF5 files

    This adapter lazily loads array data from HDF5 files using Dask. Supports reading from datasets spanning
    multiple files.
    """

    @staticmethod
    def lazy_load_hdf5_array(
        *file_paths: Union[str, Path],
        dataset: Optional[str] = None,
        swmr: bool = SWMR_DEFAULT,
        libver: str = "latest",
    ) -> dask.array.Array:
        """Lazily load arrays from possibly multiple HDF5 files and concatenate them along the first axis

        The chunks of the resulting Dask array are determined by the chunks of the constituent arrays.

        Parameters
        ----------
        file_paths : list
            A list of file paths pointing to the HDF5 files
        dataset : str
            The dataset to read from the files, for example, "/path/to/dataset" within the file
        swmr : bool
            Whether to open the files in single-writer multiple-reader mode
        libver : str
            The HDF5 library version to use
        """

        # Define helper functions for reading and getting specs of HDF5 arrays with dask.delayed
        def _read_hdf5_array(fpath: Union[str, Path]) -> NDArray[Any]:
            f = h5py.File(fpath, "r", swmr=swmr, libver=libver)
            return f[dataset] if dataset else f

        def _get_hdf5_specs(
            fpath: Union[str, Path]
        ) -> Tuple[Tuple[int, ...], Union[Tuple[int, ...], None], numpy.dtype]:
            with h5open(fpath, dataset, swmr=swmr, libver=libver) as ds:
                result = ds.shape, ds.chunks, ds.dtype
            return result

        # Need to know shapes/dtypes of constituent arrays to load them lazily
        shapes_chunks_dtypes = [_get_hdf5_specs(fpath) for fpath in file_paths]
        dtype = shapes_chunks_dtypes[0][2]
        if dtype == numpy.dtype("O"):
            # TODO: It should be possible to put this in dask.delayed too -- needs to be thoroughly tested
            warnings.warn(
                f"The dataset {dataset} is of object type, using a "
                "Python-only feature of h5py that is not supported by "
                "HDF5 in general. Read more about that feature at "
                "https://docs.h5py.org/en/stable/special.html. "
                "Consider using a fixed-length field instead. "
                "Tiled will serve an empty placeholder, unless the "
                "object is of size 1, where it will attempt to repackage "
                "the data into a numpy array."
            )

            check_str_dtype = h5py.check_string_dtype(dtype)
            if check_str_dtype.length is None:
                # TODO: refactor and test
                with h5open(
                    file_paths[0], dataset=dataset, swmr=swmr, libver=libver
                ) as value:
                    dataset_names = value.file[value.file.name + "/" + dataset][...][()]
                    if value.size == 1:
                        arr = dask.array.from_array(numpy.array(dataset_names))
                    else:
                        arr = dask.array.empty(shape=())
                return arr
            return dask.array.empty(shape=())

        if not any([shape for shape, _, _ in shapes_chunks_dtypes]):
            # All shapes are empty -> all arrays are zero-dimensional (scalars)
            array = dask.array.stack([_read_hdf5_array(fp)[()] for fp in file_paths])
        else:
            # Use delayed loading to read the arrays from the files
            delayed = [dask.delayed(_read_hdf5_array)(fpath) for fpath in file_paths]
            arrs = [
                dask.array.from_delayed(val, shape=shape, dtype=dtype).rechunk(
                    chunks=chunk_shape or "auto"
                )
                for (val, (shape, chunk_shape, dtype)) in zip(
                    delayed, shapes_chunks_dtypes
                )
            ]
            array = dask.array.concatenate(arrs, axis=0) if len(arrs) > 1 else arrs[0]

        return array

    @classmethod
    def from_catalog(
        cls,
        data_source: DataSource[ArrayStructure],
        node: Node,
        /,
        dataset: Optional[str] = None,
        slice: Optional[Union[str, NDSlice]] = None,
        squeeze: Optional[bool] = False,
        swmr: bool = SWMR_DEFAULT,
        libver: str = "latest",
        **kwargs: Optional[Any],
    ) -> "HDF5ArrayAdapter":
        structure = data_source.structure
        assets = data_source.assets
        data_uris = [
            ast.data_uri for ast in assets if ast.parameter == "data_uris"
        ] or [assets[0].data_uri]
        file_paths = [path_from_uri(uri) for uri in data_uris]

        array = cls.lazy_load_hdf5_array(
            *file_paths, dataset=dataset, swmr=swmr, libver=libver
        )

        if slice:
            if isinstance(slice, str):
                slice = NDSlice.from_numpy_str(slice)
            array = array[slice]
        if squeeze:
            array = array.squeeze()

        if array.shape != tuple(structure.shape):
            raise ValueError(
                f"Shape mismatch between array data and structure: "
                f"{array.shape} != {tuple(structure.shape)}"
            )
        if array.dtype != structure.data_type.to_numpy_dtype():
            raise ValueError(
                f"Data type mismatch between array data and structure: "
                f"{array.dtype} != {structure.data_type.to_numpy_dtype()}"
            )

        # TODO: Possibly rechunk according to structure.chunks? Is it expensive/necessary?
        # array = dask.array.rechunk(array, chunks=structure.chunks)

        # Pull additional metadata from the file attributes
        metadata = copy.deepcopy(node.metadata_)
        metadata.update(
            get_hdf5_attrs(data_uris[0], dataset, swmr=swmr, libver=libver, **kwargs)
        )

        return cls(
            array,
            structure,
            metadata=metadata,
            specs=node.specs,
        )

    @classmethod
    def from_uris(
        cls,
        *data_uris: str,
        dataset: Optional[str] = None,
        slice: Optional[Union[str, NDSlice]] = None,
        squeeze: bool = False,
        swmr: bool = SWMR_DEFAULT,
        libver: str = "latest",
        **kwargs: Optional[Any],
    ) -> "HDF5ArrayAdapter":
        file_paths = [path_from_uri(uri) for uri in data_uris]

        array = cls.lazy_load_hdf5_array(
            *file_paths, dataset=dataset, swmr=swmr, libver=libver
        )

        # Apply slice and squeeze operations, if specified
        if slice:
            if isinstance(slice, str):
                slice = NDSlice.from_numpy_str(slice)
            array = array[slice]
        if squeeze:
            array = array.squeeze()

        # Construct the structure and pull additional metadata from the file attributes
        structure = ArrayStructure.from_array(array)
        metadata = get_hdf5_attrs(
            data_uris[0], dataset, swmr=swmr, libver=libver, **kwargs
        )

        return cls(array, structure, metadata=metadata)


class HDF5Adapter(Mapping[str, Union["HDF5Adapter", HDF5ArrayAdapter]], IndexersMixin):
    """Adapter for HDF5 files

    This map the structure of an HDF5 file onto a "Tree" of array structures.

    Parameters
    ----------
    tree : dict
        A dictionary representing the HDF5 file or group. The keys are the names of the groups or datasets,
        and the values are either dictionaries (representing groups) or None (representing datasets).
        HDF5 datasets will be mapped to HDF5ArrayAdapter instances, and groups will be mapped to HDF5Adapter
        instances. The tree is rooted at the 'dataset' node.
    data_uris : str
        The URI of the file, or a list of URIs if the dataset spans multiple files.
    dataset : str
        The dataset to read, for example, "/path/to/dataset" within the file. If supplied, this path will
        effectively become the root of the adapter.
    metadata : dict
        Metadata for the adapter
    specs : list
        A list of specs for the adapter
    kwargs : dict
        Additional keyword arguments, such as swmr, libver, etc. -- they are not stored as separate attributes

    Examples
    --------

    From the root node of a file given a filepath

    >>> import h5py
    >>> HDF5Adapter.from_uri("file://localhost/path/to/file.h5")

    """

    structure_family = StructureFamily.container

    def __init__(
        self,
        tree: Union[dict[str, Any], Sentinel],
        *data_uris: str,
        dataset: Optional[str] = None,
        structure: Optional[ArrayStructure] = None,
        metadata: Optional[JSON] = None,
        specs: Optional[List[Spec]] = None,
        **kwargs: Optional[Any],
    ) -> None:
        if tree == HDF5_BROKEN_LINK:
            raise BrokenLink(
                f"Unable to open object at {data_uris[0]}"
                + (f"/{dataset}" if dataset else "")
            )
        self._tree: dict[str, Any] = tree  # type: ignore
        self.uris = data_uris
        self.dataset = dataset  # Referenced to the root of the file
        self.specs = specs or []
        self._metadata = metadata or {}
        self._kwargs = kwargs  # e.g. swmr, libver, etc.

    @classmethod
    def from_catalog(
        cls,
        # An HDF5 node may reference a dataset (array) or group (container).
        data_source: DataSource[Union[ArrayStructure, None]],
        node: Node,
        /,
        dataset: Optional[Union[str, list[str]]] = None,
        swmr: bool = SWMR_DEFAULT,
        libver: str = "latest",
        **kwargs: Optional[Any],
    ) -> Union["HDF5Adapter", HDF5ArrayAdapter]:
        # Convert the dataset representation (for backward compatibility)
        dataset = dataset or kwargs.get("path") or []
        if not isinstance(dataset, str):
            dataset = "/".join(dataset)

        # If the data source is an array, return an HDF5ArrayAdapter
        if data_source.structure_family == StructureFamily.array:
            return HDF5ArrayAdapter.from_catalog(
                data_source,  # type: ignore
                node,
                dataset=dataset,
                swmr=swmr,
                libver=libver,
                **kwargs,
            )

        # Initialize adapter for the entire HDF5 tree
        # If multiple data_uri assets are supplied, only the first one is traversed, but the rest of the uris
        # will be kept in case we need to read an array that spans all of them.
        assets = data_source.assets
        assert len(assets) > 0, "No assets found in data source"
        data_uris = [
            ast.data_uri for ast in assets if ast.parameter == "data_uris"
        ] or [assets[0].data_uri]
        file_path = path_from_uri(data_uris[0])
        with h5open(file_path, dataset, swmr=swmr, libver=libver) as file:
            tree = parse_hdf5_tree(file)

        if tree == HDF5_DATASET:
            raise ValueError(
                "Erroneous structure (container) of a DataSource pointing to an HDF5 Dataset (array)."
            )

        return cls(
            tree,
            *data_uris,
            dataset=dataset,
            structure=data_source.structure,
            metadata=node.metadata_,
            specs=node.specs,
            swmr=swmr,
            libver=libver,
            **kwargs,
        )

    @classmethod
    def from_uris(
        cls,
        *data_uris: str,
        dataset: Optional[str] = None,
        swmr: bool = SWMR_DEFAULT,
        libver: str = "latest",
        **kwargs: Optional[Any],
    ) -> Union["HDF5Adapter", HDF5ArrayAdapter]:
        fpath = path_from_uri(data_uris[0])
        with h5open(fpath, dataset, swmr=swmr, libver=libver) as file:
            tree = parse_hdf5_tree(file)

        if tree == HDF5_DATASET:
            return HDF5ArrayAdapter.from_uris(
                *data_uris, dataset=dataset, swmr=swmr, libver=libver, **kwargs  # type: ignore
            )

        return cls(
            tree, *data_uris, dataset=dataset, swmr=swmr, libver=libver, **kwargs
        )

    def __repr__(self) -> str:
        return node_repr(self, list(self))

    def structure(self) -> None:
        return None

    def metadata(self) -> JSON:
        d = get_hdf5_attrs(self.uris[0], self.dataset)
        d.update(self._metadata)
        return d

    def __iter__(self) -> Iterator[Any]:
        """Iterate over the keys of the tree"""
        yield from self._tree

    def __getitem__(self, key: str) -> Union["HDF5Adapter", HDF5ArrayAdapter]:
        dataset = f"{self.dataset or ''}/{key.strip('/')}"  # Referenced to the root of the file
        node = copy.deepcopy(self._tree)
        for segment in key.strip("/").split("/"):
            if segment not in node:
                raise NoEntry(
                    f"Can not access dataset {dataset} in {self.uris[0]}: {key} not found"
                )
            node = node[segment]
            if node == HDF5_BROKEN_LINK:  # type: ignore
                raise BrokenLink(f"Unable to open object at {self.uris[0]}/{dataset}")
        if isinstance(node, dict):
            # It is an HDF5 group
            return HDF5Adapter(
                node,
                *self.uris,
                dataset=dataset,
                metadata=self._metadata,
                specs=self.specs,
                **self._kwargs,
            )
        else:
            # It is an HDF5 dataset
            return HDF5ArrayAdapter.from_uris(
                *self.uris, dataset=dataset, **self._kwargs
            )

    def get(self, key: str, *args: Any) -> Union["HDF5Adapter", HDF5ArrayAdapter]:
        """Overwrite to always raise KeyErrors for broken links and missing items"""
        return self[key]

    def __len__(self) -> int:
        return len(self._tree)

    def keys(self) -> KeysView:  # type: ignore
        return KeysView(lambda: len(self), self._keys_slice)

    def values(self) -> ValuesView:  # type: ignore
        return ValuesView(lambda: len(self), self._items_slice)

    def items(self) -> ItemsView:  # type: ignore
        return ItemsView(lambda: len(self), self._items_slice)

    def search(self, query: Any) -> None:
        raise NotImplementedError

    def read(self, fields: Optional[str] = None) -> "HDF5Adapter":
        if fields is not None:
            raise NotImplementedError
        return self

    # The following two methods are used by keys(), values(), items().

    def _keys_slice(
        self, start: int, stop: int, direction: int, page_size: Optional[int] = None
    ) -> List[Any]:
        keys = list(self._tree.keys())
        if direction < 0:
            keys = list(reversed(keys))
        return keys[start:stop]

    def _items_slice(
        self, start: int, stop: int, direction: int, page_size: Optional[int] = None
    ) -> List[Tuple[Any, Any]]:
        """

        Parameters
        ----------
        start :
        stop :
        direction :

        Returns
        -------

        """
        items = [(key, self[key]) for key in list(self)]
        if direction < 0:
            items = list(reversed(items))
        return items[start:stop]

    def inlined_contents_enabled(self, depth: int) -> bool:
        return depth <= INLINED_DEPTH



import time
from copy import copy, deepcopy
from dataclasses import asdict
from pathlib import Path
from typing import Dict, List, Union
from urllib.parse import parse_qs, urlparse

import json_merge_patch
import jsonpatch
import orjson
from httpx import URL

from tiled.client.context import Context

from ..structures.core import STRUCTURE_TYPES, Spec, StructureFamily
from ..structures.data_source import DataSource
from ..utils import UNCHANGED, DictView, ListView, patch_mimetypes, safe_json_dump
from .metadata_update import apply_update_patch
from .utils import MSGPACK_MIME_TYPE, handle_error, normalize_specs, retry_context

# TODO: Duplicated from  tiled.type_aliases to prevent importing numpy
# After #1407 replace AnyAdapter with the BaseClass and remove this redefinition
JSON_ITEM = Union[str, int, float, bool, Dict[str, "JSON_ITEM"], List["JSON_ITEM"]]


class MetadataRevisions:
    def __init__(self, context, link):
        self._cached_len = None
        self.context = context
        self._link = link

    def __len__(self):
        LENGTH_CACHE_TTL = 1  # second

        now = time.monotonic()
        if self._cached_len is not None:
            length, deadline = self._cached_len
            if now < deadline:
                # Used the cached value and do not make any request.
                return length

        for attempt in retry_context():
            with attempt:
                content = handle_error(
                    self.context.http_client.get(
                        self._link,
                        headers={"Accept": MSGPACK_MIME_TYPE},
                        params={
                            **parse_qs(urlparse(self._link).query),
                            "page[offset]": 0,
                            "page[limit]": 0,
                        },
                    )
                ).json()
        length = content["meta"]["count"]
        self._cached_len = (length, now + LENGTH_CACHE_TTL)
        return length

    def __getitem__(self, item_):
        self._cached_len = None

        if isinstance(item_, int):
            offset = item_
            limit = 1

            for attempt in retry_context():
                with attempt:
                    content = handle_error(
                        self.context.http_client.get(
                            self._link,
                            headers={"Accept": MSGPACK_MIME_TYPE},
                            params={
                                **parse_qs(urlparse(self._link).query),
                                "page[offset]": offset,
                                "page[limit]": limit,
                            },
                        )
                    ).json()
            (result,) = content["data"]
            return result

        elif isinstance(item_, slice):
            offset = item_.start
            if offset is None:
                offset = 0
            if item_.stop is None:
                params = f"?page[offset]={offset}"
            else:
                limit = item_.stop - offset
                params = f"?page[offset]={offset}&page[limit]={limit}"

            next_page_url = self._link + params
            result = []
            while next_page_url is not None:
                for attempt in retry_context():
                    with attempt:
                        content = handle_error(
                            self.context.http_client.get(
                                next_page_url, headers={"Accept": MSGPACK_MIME_TYPE}
                            )
                        ).json()
                if len(result) == 0:
                    result = content.copy()
                else:
                    result["data"].append(content["data"])
                next_page_url = content["links"]["next"]

            return result["data"]

    def delete_revision(self, n):
        for attempt in retry_context():
            with attempt:
                handle_error(
                    self.context.http_client.delete(
                        self._link,
                        params={**parse_qs(urlparse(self._link).query), "number": n},
                    )
                )


class BaseClient:
    # The HTTP spec does not define a size limit for URIs,
    # but a common setting is 4K or 8K (for all the headers together).
    # As another reference point, Internet Explorer imposes a
    # 2048-character limit on URLs.
    URL_CHARACTER_LIMIT = 2_000  # number of characters

    def __init__(
        self,
        context: Context,
        *,
        item,
        structure_clients,
        structure=None,
        include_data_sources=False,
    ):
        self._context = context
        self._item = item
        self._cached_len = None  # a cache just for __len__
        self.structure_clients = structure_clients
        self._metadata_revisions = None
        self._include_data_sources = include_data_sources
        attributes = self.item["attributes"]
        structure_family = attributes["structure_family"]

        if structure is not None:
            # Allow the caller to optionally hand us a structure that is already
            # parsed from a dict into a structure dataclass.
            self._structure = structure
        elif structure_family == StructureFamily.container:
            self._structure = None
        else:
            structure_type = STRUCTURE_TYPES[attributes["structure_family"]]
            self._structure = structure_type.from_json(attributes["structure"])

        super().__init__()

    def structure(self):
        """
        Return a dataclass describing the structure of the data.
        """
        if getattr(self._structure, "resizable", None):
            # In the future, conditionally fetch updated information.
            raise NotImplementedError(
                "The server has indicated that this has a dynamic, resizable "
                "structure and this version of the Tiled Python client cannot "
                "cope with that."
            )
        return self._structure

    def login(self):
        """
        Depending on the server's authentication method, this will prompt for username/password:

        >>> c.login()
        Username: USERNAME
        Password: <input is hidden>

        or prompt you to open a link in a web browser to login with a third party:

        >>> c.login()
        You have ... minutes visit this URL

        https://...

        and enter the code: XXXX-XXXX
        """
        self.context.authenticate()

    def logout(self):
        """
        Log out.

        This method is idempotent: if you are already logged out, it will do nothing.
        """
        self.context.logout()

    def __repr__(self):
        return f"<{type(self).__name__}>"

    @property
    def context(self):
        return self._context

    def refresh(self):
        params = {
            **parse_qs(urlparse(self.uri).query),
        }
        if self._include_data_sources:
            params["include_data_sources"] = self._include_data_sources
        for attempt in retry_context():
            with attempt:
                content = handle_error(
                    self.context.http_client.get(
                        self.uri,
                        headers={"Accept": MSGPACK_MIME_TYPE},
                        params=params,
                    )
                ).json()
        self._item = content["data"]
        if self.structure_family != StructureFamily.container:
            structure_type = STRUCTURE_TYPES[self.structure_family]
            self._structure = structure_type.from_json(
                self._item["attributes"]["structure"]
            )
        return self

    @property
    def item(self):
        "JSON payload describing this item. Mostly for internal use."
        return self._item

    @property
    def metadata(self) -> DictView[str, JSON_ITEM]:
        "Metadata about this data source."
        # Ensure this is immutable (at the top level) to help the user avoid
        # getting the wrong impression that editing this would update anything
        # persistent.
        return DictView(self._item["attributes"]["metadata"])

    @property
    def parent(self):
        "Returns a client for the parent of this node."
        # this import takes about 230 ns.
        from .constructors import from_context

        return from_context(
            context=self.context,
            structure_clients=self.structure_clients,
            node_path_parts=self._item["attributes"]["ancestors"],
            include_data_sources=self._include_data_sources,
        )

    def metadata_copy(self):
        """
        Generate a mutable copy of metadata, specs, and access_tags for
        validating metadata (useful with update_metadata())
        """
        metadata = deepcopy(self._item["attributes"]["metadata"])
        specs = [Spec(**spec) for spec in self._item["attributes"]["specs"]]
        access_tags = deepcopy(self._item["attributes"]["access_blob"].get("tags", []))
        return [
            md for md in [metadata, specs, access_tags] if md is not None
        ]  # returning as list of mutable items

    @property
    def specs(self) -> ListView[Spec]:
        "List of specifications describing the structure of the metadata and/or data."
        return ListView([Spec(**spec) for spec in self._item["attributes"]["specs"]])

    @property
    def access_blob(self) -> DictView[str, JSON_ITEM]:
        "Authorization information about this node, in blob form"
        access_blob = self._item["attributes"]["access_blob"]
        if access_blob is None:
            raise AttributeError("Node has no attribute 'access_blob'")
        # Ensure this is immutable (at the top level) to help the user avoid
        # getting the wrong impression that editing this would update anything
        # persistent.
        return DictView(access_blob)

    @property
    def uri(self):
        "Direct link to this entry"
        return self.item["links"]["self"]

    @property
    def structure_family(self):
        "Quick access to this entry"
        return StructureFamily[self.item["attributes"]["structure_family"]]

    def data_sources(self):
        data_sources_json = (
            self.include_data_sources().item["attributes"].get("data_sources")
        )
        if data_sources_json is None:
            return None
        return [DataSource.from_json(d) for d in data_sources_json]

    def include_data_sources(self):
        """
        Ensure that data source and asset information is fetched.

        If it has already been fetched, this is a no-op.
        """
        if self._include_data_sources:
            return self  # no op
        return self.new_variation(include_data_sources=True).refresh()

    def new_variation(
        self,
        structure_clients=UNCHANGED,
        include_data_sources=UNCHANGED,
        structure=UNCHANGED,
        **kwargs,
    ):
        """
        This is intended primarily for internal use and use by subclasses.
        """
        if structure_clients is UNCHANGED:
            structure_clients = self.structure_clients
        if include_data_sources is UNCHANGED:
            include_data_sources = self._include_data_sources
        if structure is UNCHANGED:
            structure = self._structure
        return type(self)(
            self.context,
            item=self._item,
            structure=structure,
            structure_clients=structure_clients,
            include_data_sources=include_data_sources,
            **kwargs,
        )

    def asset_manifest(self, data_sources):
        """
        Return a manifest of the relative paths of the contents in each asset.

        This return a dictionary keyed on asset ID.
        Assets backed by a single file are mapped to None (no manifest).
        Asset backed by a directory of files are mapped to a list of relative paths.

        Parameters
        ----------
        data_sources : dict
            The value returned by ``.data_sources()``. This is passed in explicitly
            to avoid fetching it twice in common usages. It also enables passing in
            a subset of the data_sources of interest.
        """
        manifests = {}
        for data_source in data_sources:
            manifest_link = self.item["links"]["self"].replace(
                "/metadata", "/asset/manifest", 1
            )
            for asset in data_source.assets:
                if asset.is_directory:
                    for attempt in retry_context():
                        with attempt:
                            manifest = handle_error(
                                self.context.http_client.get(
                                    manifest_link,
                                    params={
                                        **parse_qs(urlparse(manifest_link).query),
                                        "id": asset.id,
                                    },
                                )
                            ).json()["manifest"]
                else:
                    manifest = None
                manifests[asset.id] = manifest
        return manifests

    def raw_export(self, destination_directory=None, max_workers=4):
        """
        Download the raw assets backing this node.

        This may produce a single file or a directory.

        Parameters
        ----------
        destination_directory : Path, optional
            Destination for downloaded assets. Default is current working directory
        max_workers : int, optional
            Number of parallel workers downloading data. Default is 4.

        Returns
        -------
        paths : List[Path]
            Filepaths of exported files
        """
        if destination_directory is None:
            destination_directory = Path.cwd()
        else:
            destination_directory = Path(destination_directory)

        # Import here to defer the import of rich (for progress bar).
        from .download import ATTACHMENT_FILENAME_PLACEHOLDER, download

        urls = []
        paths = []
        data_sources = self.include_data_sources().data_sources()
        asset_manifest = self.asset_manifest(data_sources)
        if len(data_sources) != 1:
            raise NotImplementedError(
                "Export of multiple data sources not yet supported"
            )
        for data_source in data_sources:
            bytes_link = self.item["links"]["self"].replace(
                "/metadata", "/asset/bytes", 1
            )
            for asset in data_source.assets:
                if len(data_source.assets) == 1:
                    # Only one asset: keep the name simple.
                    base_path = destination_directory
                else:
                    # Multiple assets: Add a subdirectory named for the asset
                    # id to namespace each asset.
                    base_path = Path(destination_directory, str(asset.id))
                if asset.is_directory:
                    relative_paths = asset_manifest[asset.id]
                    urls.extend(
                        [
                            URL(
                                bytes_link,
                                params={
                                    **parse_qs(urlparse(bytes_link).query),
                                    "id": asset.id,
                                    "relative_path": relative_path,
                                },
                            )
                            for relative_path in relative_paths
                        ]
                    )
                    paths.extend(
                        [
                            Path(base_path, relative_path)
                            for relative_path in relative_paths
                        ]
                    )
                else:
                    urls.append(
                        URL(
                            bytes_link,
                            params={
                                **parse_qs(urlparse(bytes_link).query),
                                "id": asset.id,
                            },
                        )
                    )
                    paths.append(Path(base_path, ATTACHMENT_FILENAME_PLACEHOLDER))
        return download(self.context.http_client, urls, paths, max_workers=max_workers)

    @property
    def formats(self):
        "List formats that the server can export this data as."
        formats = set()
        for spec in self.item["attributes"]["specs"]:
            formats.update(self.context.server_info.formats.get(spec["name"], []))
        formats.update(
            self.context.server_info.formats[
                self.item["attributes"]["structure_family"]
            ]
        )
        return sorted(formats)

    def update_metadata(
        self, metadata=None, specs=None, access_tags=None, *, drop_revision=False
    ):
        """
        EXPERIMENTAL: Update metadata via a `dict.update`- like interface.

        `update_metadata` is a user-friendly wrapper for `patch_metadata`.
        This is subject to change or removal without notice.

        Parameters
        ----------
        metadata : dict, optional
            User metadata. May be nested. Must contain only basic types
            (e.g. numbers, strings, lists, dicts) that are JSON-serializable.
        specs : List[str], optional
            List of names that are used to label that the data and/or metadata
            conform to some named standard specification.
        access_tags: List[str], optional
            Server-specific authZ tags in list form, used to confer access to the node.
        drop_revision : bool, optional
            Replace current version without saving current version as a revision.
            Use with caution.

        See Also
        --------
        patch_metadata
        replace_metadata

        Notes
        -----
        `update_metadata` constructs a JSON Patch (RFC6902) by comparing user updates
        to existing metadata. It uses a slight variation of JSON Merge Patch (RFC7386)
        as an intermediary to implement a python `dict.update`-like user-friendly
        interface, but with additional features like key deletion (see examples) and
        support for `None (null)` values.

        Examples
        --------

        Add or update a key-value pair at the top or a nested level

        >>> node.update_metadata({'key': new_value})
        >>> node.update_metadata({'top_key': {'nested_key': new_value}})

        Remove an existing key

        >>> from tiled.client.metadata_update import DELETE_KEY
        >>> node.update_metadata({'key_to_be_deleted': DELETE_KEY})

        Interactively update complex metadata using a copy of original structure
        (e.g., in iPython you may use tab completion to navigate nested metadata)

        >>> md = node.metadata_copy()[0]
        >>> md['L1_key']['L2_key']['L3_key'] = new_value  # use tab completion
        >>> md['unwanted_key'] = DELETE_KEY
        >>> node.update_metadata(metadata=md)  # Update the copy on the server
        """
        metadata_patch, specs_patch, access_blob_patch = self.build_metadata_patches(
            metadata=metadata,
            specs=specs,
            access_tags=access_tags,
        )
        self.patch_metadata(
            metadata_patch=metadata_patch,
            specs_patch=specs_patch,
            access_blob_patch=access_blob_patch,
            drop_revision=drop_revision,
        )

    def build_metadata_patches(self, metadata=None, specs=None, access_tags=None):
        """
        Build valid JSON Patches (RFC6902) for metadata and metadata validation
        specs accepted by `patch_metadata`.

        Parameters
        ----------
        metadata : dict, optional
            User metadata. May be nested. Must contain only basic types
            (e.g. numbers, strings, lists, dicts) that are JSON-serializable.

        specs : list[Spec], optional
            Metadata validation specifications.

        access_tags: List[str], optional
            Server-specific authZ tags in list form, used to confer access to the node.

        Returns
        -------
        metadata_patch : list[dict]
            A JSON serializable object representing a valid JSON patch (RFC6902)
            for metadata.
        specs_patch : list[dict]
            A JSON serializable object representing a valid JSON patch (RFC6902)
            for metadata validation specifications.
        access_blob_patch : list[dict]
            A JSON serializable object representing a valid JSON patch (RFC6902)
            for access control fields that are stored in the access_blob.

        See Also
        --------
        patch_metadata
        update_metadata

        Notes
        -----
        `build_metadata_patch` constructs a JSON Patch (RFC6902) by comparing user updates
        to existing metadata/specs. It uses a slight variation of JSON Merge Patch (RFC7386)
        as an intermediary to implement a python `dict.update`-like user-friendly
        interface, but with additional features like key deletion (see examples) and
        support for `None (null)` values.

        Examples
        --------

        Build a patch for adding/updating a key-value pair at the top or a nested level

        >>> patches = node.build_metadata_patches({'key': new_value})
        >>> patches = node.build_metadata_patches({'top_key': {'nested_key': new_value}})

        Build patches for metadata and specs ("mp", "sp")

        >>> mp, sp = node.build_metadata_patches(metadata=metadata, specs=specs)

        Build a patch for removing an existing key

        >>> from tiled.client.metadata_update import DELETE_KEY
        >>> node.build_metadata_patches({'key_to_be_deleted': DELETE_KEY})

        Interactively build a patch for complex metadata (e.g., in iPython you may use
        tab completion to navigate nested metadata)

        >>> md = node.metadata_copy()[0]
        >>> md['L1_key']['L2_key']['L3_key'] = new_value  # use tab completion
        >>> md['unwanted_key'] = DELETE_KEY
        >>> node.build_metadata_patches(metadata=md)  # Generate the patch
        """

        if metadata is None:
            metadata_patch = []
        else:
            md_copy = deepcopy(self._item["attributes"]["metadata"])
            metadata_patch = jsonpatch.JsonPatch.from_diff(
                self._item["attributes"]["metadata"],
                apply_update_patch(md_copy, metadata),
                dumps=orjson.dumps,
            ).patch

        if specs is None:
            specs_patch = None
        else:
            sp_copy = [spec["name"] for spec in self._item["attributes"]["specs"]]
            specs_patch = (
                []
                if specs is None
                else jsonpatch.JsonPatch.from_diff(
                    sp_copy, specs, dumps=orjson.dumps
                ).patch
            )

        if not access_tags:
            # empty list of access_tags should be a no-op
            access_blob_patch = None
        else:
            ab_copy = deepcopy(self._item["attributes"]["access_blob"])
            access_blob = {"tags": access_tags}
            access_blob_patch = jsonpatch.JsonPatch.from_diff(
                self._item["attributes"]["access_blob"],
                apply_update_patch(ab_copy, access_blob),
                dumps=orjson.dumps,
            ).patch

        return metadata_patch, specs_patch, access_blob_patch

    def _build_json_patch(self, origin, update_patch):
        """
        Lower level method to construct a JSON patch from an origin and update_patch.
        An "update_patch" is a `dict.update`-like specification that may include
        `DELETE_KEY` for marking a dictionary key for deletion.
        """
        if update_patch is None:
            return []
        patch = jsonpatch.JsonPatch.from_diff(
            origin, apply_update_patch(origin, update_patch), dumps=orjson.dumps
        )
        return patch.patch

    def _build_metadata_revisions(self):
        if self._metadata_revisions is None:
            link = self.item["links"]["self"].replace("/metadata", "/revisions", 1)
            self._metadata_revisions = MetadataRevisions(self.context, link)

        return self._metadata_revisions

    def patch_metadata(
        self,
        metadata_patch=None,
        specs_patch=None,
        access_blob_patch=None,
        content_type=patch_mimetypes.JSON_PATCH,
        drop_revision=False,
    ):
        """
        EXPERIMENTAL: Patch metadata using a JSON Patch (RFC6902).

        This is subject to change or removal without notice.

        Parameters
        ----------
        metadata_patch : List[dict], optional
            JSON-serializable patch to be applied to metadata
        specs_patch : List[dict], optional
            JSON-serializable patch to be applied to metadata validation
            specifications list
        access_blob_patch : List[dict], optional
            JSON-serializable patch to be applied to the access_blob
        content_type : str
            Mimetype of the patches. Acceptable values are:

            * "application/json-patch+json"
              (See https://datatracker.ietf.org/doc/html/rfc6902)
            * "application/merge-patch+json"
              (See https://datatracker.ietf.org/doc/html/rfc7386)
        drop_revision : bool, optional
            Replace current version without saving current version as a revision.
            Use with caution.

        See Also
        --------
        update_metadata
        replace_metadata
        """

        self._cached_len = None

        def patcher(doc, patch, patch_type):
            # this helper function applies a given type of patch to the document
            # and returns the modified document
            if patch_type == patch_mimetypes.JSON_PATCH:
                return jsonpatch.apply_patch(
                    doc=doc,
                    patch=patch,
                    in_place=False,
                )
            if patch_type == patch_mimetypes.MERGE_PATCH:
                return json_merge_patch.merge(doc, patch)
            raise ValueError(
                f"Unsupported patch type {content_type}. "
                f"Acceptable values are: {', '.join(patch_mimetypes)}."
            )

        assert content_type in patch_mimetypes
        if specs_patch is None:
            normalized_specs_patch = None
        else:
            normalized_specs_patch = []

        if content_type == patch_mimetypes.JSON_PATCH:
            if specs_patch:
                for spec_patch in copy(specs_patch):
                    value = spec_patch.get("value", None)
                    if isinstance(value, str):
                        spec_patch["value"] = asdict(Spec(value))
                    normalized_specs_patch.append(spec_patch)
        elif content_type == patch_mimetypes.MERGE_PATCH:
            if specs_patch:
                for spec in specs_patch:
                    if isinstance(spec, str):
                        spec = Spec(spec)
                    normalized_specs_patch.append(asdict(spec))

        data = {
            "content-type": content_type,
            "metadata": metadata_patch,
            "specs": normalized_specs_patch,
            "access_blob": access_blob_patch,
        }
        params = {}
        if drop_revision:
            params["drop_revision"] = True

        for attempt in retry_context():
            with attempt:
                content = handle_error(
                    self.context.http_client.patch(
                        self.item["links"]["self"],
                        content=safe_json_dump(data),
                        params=params,
                    )
                ).json()

        if metadata_patch is not None:
            if "metadata" in content:
                # Metadata was accepted and modified by the specs validator on the server side.
                # It is updated locally using the new version.
                self._item["attributes"]["metadata"] = content["metadata"]
            else:
                # Metadata was accepted as it is by the server.
                # It is updated locally with the version submitted by the client.
                self._item["attributes"]["metadata"] = patcher(
                    dict(self.metadata), metadata_patch, content_type
                )

        if specs_patch is not None:
            current_specs = self._item["attributes"]["specs"]
            patched_specs = patcher(current_specs, normalized_specs_patch, content_type)
            self._item["attributes"]["specs"] = patched_specs

        if access_blob_patch is not None:
            if "access_blob" in content:
                self._item["attributes"]["access_blob"] = content["access_blob"]
            else:
                self._item["attributes"]["access_blob"] = patcher(
                    dict(self.access_blob), access_blob_patch, content_type
                )

    def replace_metadata(
        self, metadata=None, specs=None, access_tags=None, drop_revision=False
    ):
        """
        EXPERIMENTAL: Replace metadata entirely (see update_metadata).

        This is subject to change or removal without notice.

        Parameters
        ----------
        metadata : dict, optional
            User metadata. May be nested. Must contain only basic types
            (e.g. numbers, strings, lists, dicts) that are JSON-serializable.
        specs : List[str], optional
            List of names that are used to label that the data and/or metadata
            conform to some named standard specification.
        access_tags: List[str], optional
            Server-specific authZ tags in list form, used to confer access to the node.
        drop_revision : bool, optional
            Replace current version without saving current version as a revision.
            Use with caution.

        See Also
        --------
        update_metadata
        patch_metadata
        """

        self._cached_len = None

        if access_tags is None:
            access_blob = None
        else:
            access_blob = {"tags": access_tags}

        data = {
            "metadata": metadata,
            "specs": normalize_specs(specs),
            "access_blob": access_blob,
        }
        params = {}
        if drop_revision:
            params["drop_revision"] = True

        for attempt in retry_context():
            with attempt:
                content = handle_error(
                    self.context.http_client.put(
                        self.item["links"]["self"],
                        content=safe_json_dump(data),
                        params=params,
                    )
                ).json()

        if metadata is not None:
            if "metadata" in content:
                # Metadata was accepted and modified by the specs validator on the server side.
                # It is updated locally using the new version.
                self._item["attributes"]["metadata"] = content["metadata"]
            else:
                # Metadata was accepted as it is by the server.
                # It is updated locally with the version submitted by the client.
                self._item["attributes"]["metadata"] = metadata

        if specs is not None:
            self._item["attributes"]["specs"] = normalize_specs(specs)

        if access_blob is not None:
            if "access_blob" in content:
                self._item["attributes"]["access_blob"] = content["access_blob"]
            else:
                self._item["attributes"]["access_blob"] = access_blob

    @property
    def metadata_revisions(self):
        if self._metadata_revisions is None:
            link = self.item["links"]["self"].replace("/metadata", "/revisions", 1)
            self._metadata_revisions = MetadataRevisions(self.context, link)

        return self._metadata_revisions

    def delete(self, recursive: bool = False, external_only: bool = True) -> None:
        """Delete the node and its contents, if any.

        Parameters
        ----------
        recursive : bool, optional
            If True, descend into sub-nodes and delete their contents too.
            Defaults to False.
        external_only : bool, optional
            If True, only delete externally-managed data. Defaults to True.
        """

        self._cached_len = None
        for attempt in retry_context():
            with attempt:
                handle_error(
                    self.context.http_client.delete(
                        f"{self.uri}",
                        params={"recursive": recursive, "external_only": external_only},
                    )
                )

    def close_stream(self):
        "Declare the end of a stream of writes to this node."
        endpoint = self.uri.replace("/metadata/", "/stream/close/", 1)
        for attempt in retry_context():
            with attempt:
                handle_error(self.context.http_client.delete(endpoint))

    def __dask_tokenize__(self):
        return (type(self), self.uri)



import secrets
from datetime import timedelta
from functools import cache
from typing import Any, List, Optional

from pydantic import Field
from pydantic.dataclasses import dataclass
from pydantic_settings import BaseSettings, SettingsConfigDict


# hashable cache key for use in tiled.authn_database.connection_pool
@dataclass(unsafe_hash=True)
class DatabaseSettings:
    uri: Optional[str] = None
    pool_size: int = 5
    pool_pre_ping: bool = True
    max_overflow: int = 5


class Settings(BaseSettings):
    """A BaseSettings object defining configuration for the tiled instance.
    For loading variables from the environment, prefix with TILED_ and see:
    https://docs.pydantic.dev/latest/concepts/pydantic_settings/#parsing-environment-variable-values
    """

    tree: Any = None
    allow_anonymous_access: bool = False
    allow_origins: List[str] = Field(default_factory=list)
    authenticator: Any = None
    # These 'single user' settings are only applicable if authenticator is None.
    single_user_api_key: str = secrets.token_hex(32)
    # The first key will be used for encryption. Each key will be tried in turn for decryption.
    secret_keys: List[str] = [secrets.token_hex(32)]
    access_token_max_age: timedelta = timedelta(minutes=15)
    refresh_token_max_age: timedelta = timedelta(days=7)
    session_max_age: timedelta = timedelta(days=365)
    # Put a fairly low limit on the maximum size of one chunk, keeping in mind
    # that data should generally be chunked. When we implement async responses,
    # we can raise this global limit.
    response_bytesize_limit: int = 300_000_000  # 300 MB
    # The largest number of items in a container for which the metadata endpoint
    # will return the exact count; this becomes the lower bound on the stimate if
    # an approximate number can not be obtained otherwise.
    exact_count_limit: int = 100
    reject_undeclared_specs: bool = False
    # "env_prefix does not apply to fields with alias"
    # https://docs.pydantic.dev/latest/concepts/pydantic_settings/#environment-variable-names
    database_settings: DatabaseSettings = Field(
        DatabaseSettings(), validation_alias="TILED_DATABASE"
    )
    # Connection pool configurations for catalog and storage DBs
    catalog_pool_size: int = 5
    storage_pool_size: int = 5
    catalog_max_overflow: int = 10
    storage_max_overflow: int = 10
    database_init_if_not_exists: bool = False
    expose_raw_assets: bool = True

    model_config = SettingsConfigDict(
        env_prefix="TILED_",
        nested_model_default_partial_update=True,
        env_nested_delimiter="_",
    )


@cache
def get_settings() -> Settings:
    return Settings()



import contextlib
import sqlite3
import sys
import tempfile
import threading
import time
import uuid
from enum import IntEnum
from pathlib import Path

import httpx
import pytest
import uvicorn
from sqlalchemy import text
from sqlalchemy.engine import Connection
from sqlalchemy.ext.asyncio import create_async_engine

from ..client import context
from ..client.base import BaseClient
from ..utils import ensure_specified_sql_driver

if sys.version_info < (3, 9):
    import importlib_resources as resources
else:
    from importlib import resources  # Python >= 3.9 only


@contextlib.contextmanager
def fail_with_status_code(status_code):
    with pytest.raises(httpx.HTTPStatusError) as info:
        yield info
    assert info.value.response.status_code == status_code


@contextlib.asynccontextmanager
async def temp_postgres(uri):
    if uri.endswith("/"):
        uri = uri[:-1]
    # Create a fresh database.
    engine = create_async_engine(ensure_specified_sql_driver(uri))
    database_name = f"tiled_test_disposable_{uuid.uuid4().hex}"
    async with engine.connect() as connection:
        await connection.execute(
            text("COMMIT")
        )  # close the automatically-started transaction
        await connection.execute(text(f"CREATE DATABASE {database_name};"))
        await connection.commit()
    yield f"{uri}/{database_name}"
    # Drop the database.
    async with engine.connect() as connection:
        await connection.execute(
            text("COMMIT")
        )  # close the automatically-started transaction
        await connection.execute(text(f"DROP DATABASE {database_name};"))
        await connection.commit()


@contextlib.contextmanager
def enter_username_password(username, password):
    """
    Override getpass, when prompt_for_credentials with username only
    used like:

    >>> with enter_username_password(...):
    ...     # Run code that calls prompt_for_credentials and subsequently getpass.getpass().
    """

    original_username_input = context.username_input
    context.username_input = lambda: username
    original_password_input = context.password_input
    context.password_input = lambda: password
    try:
        # Ensures that raise in calling routine does not prevent context from being exited.
        yield
    finally:
        context.username_input = original_username_input
        context.password_input = original_password_input


class URL_LIMITS(IntEnum):
    HUGE = 80_000
    DEFAULT = BaseClient.URL_CHARACTER_LIMIT
    TINY = 10


@contextlib.contextmanager
def sqlite_from_dump(filename):
    """Create a SQLite db in a temporary directory, loading a SQL script.

    SQL script should be given as a filename, assumed to be in tiled/_tests/sql/
    """
    with tempfile.TemporaryDirectory() as directory:
        database_path = Path(directory, "catalog.db")
        conn = sqlite3.connect(database_path)
        ref = resources.files("tiled._tests.sql") / filename
        with resources.as_file(ref) as path:
            conn.executescript(path.read_text())
        conn.close()
        yield database_path


class Server(uvicorn.Server):
    # https://github.com/encode/uvicorn/discussions/1103#discussioncomment-941726

    def install_signal_handlers(self):
        pass

    @contextlib.contextmanager
    def run_in_thread(self):
        thread = threading.Thread(target=self.run)
        thread.start()
        try:
            # Wait for server to start up, or raise TimeoutError.
            for _ in range(100):
                time.sleep(0.1)
                if self.started:
                    break
            else:
                raise TimeoutError("Server did not start in 10 seconds.")
            host, port = self.servers[0].sockets[0].getsockname()
            yield f"http://{host}:{port}"
        finally:
            self.should_exit = True
            thread.join()


def sql_table_exists(conn: Connection, dialect: str, table_name: str) -> bool:
    """Check if a table exists in the SQLite database."""

    # Use a dialect-specific query and parameter style
    if dialect == "sqlite":
        query = f"""
            SELECT name FROM sqlite_master
            WHERE type='table' AND name='{table_name}';
        """
    elif dialect == "duckdb":
        query = f"""
            SELECT table_name FROM information_schema.tables
            WHERE table_schema = 'main' AND table_name = '{table_name}';
        """
    elif dialect == "postgresql":
        query = f"""
            SELECT tablename FROM pg_catalog.pg_tables
            WHERE schemaname = 'public' AND tablename = '{table_name}'
        """
        # NOTE: no trailing semicolon in PostgreSQL query as it is run as a COPY
    else:
        raise ValueError(f"Unsupported database dialect: {dialect}")

    with conn.cursor() as cursor:
        cursor.execute(query)
        return cursor.fetchone() is not None



import numpy as np
import pytest

from ..adapters.array import ArrayAdapter
from ..adapters.mapping import MapAdapter
from ..client import Context, from_context
from ..client.smoke import read
from ..server.app import build_app


class Broken(Exception):
    pass


class BrokenArrayAdapter(ArrayAdapter):
    def read(self, *args, **kwargs):
        raise Broken

    def read_block(self, *args, **kwargs):
        raise Broken


@pytest.fixture(scope="module")
def context():
    mapping = {
        "A": ArrayAdapter.from_array(np.array([1, 2, 3]), metadata={"A": "a"}),
        "B": BrokenArrayAdapter.from_array(np.array([4, 5, 6]), metadata={"B": "b"}),
    }

    tree = MapAdapter(mapping)
    with Context.from_app(build_app(tree)) as context:
        yield context


def test_smoke_read_list(context):
    client = from_context(context)

    faulty_list = read(client)
    assert len(faulty_list) == 1


def test_smoke_read_raise(context):
    client = from_context(context)

    with pytest.raises(Broken):
        read(client, strict=True)



import sys

from tiled.structures.array import ArrayStructure
from tiled.structures.awkward import AwkwardStructure
from tiled.structures.sparse import SparseStructure
from tiled.structures.table import TableStructure

if sys.version_info < (3, 10):
    EllipsisType = type(Ellipsis)
else:
    from types import EllipsisType

from typing import Any, Dict, List, Set, Union

JSON = Dict[str, Union[str, int, float, bool, Dict[str, "JSON"], List["JSON"]]]

Scopes = Set[str]
Query = Any  # for now...
Filters = List[Query]

AnyStructure = Union[TableStructure, ArrayStructure, SparseStructure, AwkwardStructure]

__all__ = [
    "AnyStructure",
    "EllipsisType",
    "JSON",
    "Scopes",
    "Query",
    "Filters",
]



import json
import re
from typing import Optional, Set, Tuple, Union

import numcodecs
import orjson
import pydantic_settings
from fastapi import APIRouter, Depends, HTTPException, Request
from starlette.responses import Response
from starlette.status import HTTP_400_BAD_REQUEST, HTTP_500_INTERNAL_SERVER_ERROR

from ..structures.core import StructureFamily
from ..type_aliases import Scopes
from ..utils import ensure_awaitable
from .authentication import (
    get_current_access_tags,
    get_current_principal,
    get_current_scopes,
    get_session_state,
)
from .dependencies import get_entry, get_root_tree
from .schemas import Principal
from .utils import record_timing

ZARR_BLOCK_SIZE = 10000
ZARR_BYTE_ORDER = "C"
ZARR_CODEC_SPEC = {
    "blocksize": 0,
    "clevel": 5,
    "cname": "lz4",
    "id": "blosc",
    "shuffle": 1,
}

zarr_codec = numcodecs.get_codec(ZARR_CODEC_SPEC)


def convert_chunks_for_zarr(tiled_chunks: Tuple[Tuple[int]]):
    """Convert full tiled/dask chunk specification into zarr format

    Zarr only accepts chunks of constant size along each dimension; this function
    finds a unique representation of (possibly variable-sized chunks) internal to
    Tiled ArrayAdapter in terms of zarr blocks.

    Zarr chunks must be at least of size 1, even for zero-dimensional arrays.
    """
    return [min(ZARR_BLOCK_SIZE, max(*tc, 1)) for tc in tiled_chunks]


def get_zarr_router_v2() -> APIRouter:
    router = APIRouter()

    @router.get("{path:path}.zattrs", name="Zarr .zattrs metadata")
    @router.get("/{path:path}/.zattrs", name="Zarr .zattrs metadata")
    async def get_zarr_attrs(
        request: Request,
        path: str,
        principal: Union[Principal] = Depends(get_current_principal),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        root_tree: pydantic_settings.BaseSettings = Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
    ):
        "Return entry.metadata as Zarr attributes metadata (.zattrs)"
        entry = await get_entry(
            path,
            ["read:data", "read:metadata"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            metrics=request.state.metrics,
            structure_families={
                StructureFamily.table,
                StructureFamily.container,
                StructureFamily.array,
                StructureFamily.sparse,
            },
            access_policy=getattr(request.app.state, "access_policy", None),
        )

        return Response(
            json.dumps(entry.metadata().get("attributes", {})),
            status_code=200,
            media_type="application/json",
        )

    @router.get("{path:path}.zgroup", name="Root .zgroup metadata")
    @router.get("/{path:path}/.zgroup", name="Zarr .zgroup metadata")
    async def get_zarr_group_metadata(
        request: Request,
        path: str,
        principal: Union[Principal] = Depends(get_current_principal),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        root_tree: pydantic_settings.BaseSettings = Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
    ):
        await get_entry(
            path,
            ["read:data", "read:metadata"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            metrics=request.state.metrics,
            structure_families={
                StructureFamily.table,
                StructureFamily.container,
            },
            access_policy=getattr(request.app.state, "access_policy", None),
        )

        return Response(json.dumps({"zarr_format": 2}), status_code=200)

    @router.get("/{path:path}/.zarray", name="Zarr .zarray metadata")
    async def get_zarr_array_metadata(
        request: Request,
        path: str,
        principal: Union[Principal] = Depends(get_current_principal),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        root_tree: pydantic_settings.BaseSettings = Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
    ):
        entry = await get_entry(
            path,
            ["read:data", "read:metadata"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            metrics=request.state.metrics,
            structure_families={StructureFamily.array, StructureFamily.sparse},
            access_policy=getattr(request.app.state, "access_policy", None),
        )
        structure = entry.structure()
        try:
            zarray_spec = {
                "chunks": convert_chunks_for_zarr(structure.chunks),
                "compressor": zarr_codec.get_config(),
                "dtype": structure.data_type.to_numpy_descr(),
                "fill_value": None,
                "filters": None,
                "order": ZARR_BYTE_ORDER,
                "shape": list(structure.shape),
                "zarr_format": 2,
            }
            return Response(json.dumps(zarray_spec), status_code=200)
        except Exception as err:
            print(f"Can not create .zarray metadata, {err}")
            raise HTTPException(
                status_code=HTTP_500_INTERNAL_SERVER_ERROR, detail=err.args[0]
            )

    @router.get(
        "/{path:path}",
        name="Zarr group (directory) structure or a chunk of a zarr array",
    )
    async def get_zarr_array(
        request: Request,
        path: str,
        principal: Union[Principal] = Depends(get_current_principal),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        root_tree: pydantic_settings.BaseSettings = Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
    ):
        # If a zarr block is requested, e.g. http://localhost:8000/zarr/v2/array/0.1.2.3,
        # extract it from last part of the path; use the remaining path to get the entry.
        zarr_block_indx = None
        if block := path.strip("/").split("/")[-1]:
            if re.fullmatch(r"^(?:\d+\.)*\d+$", block):
                zarr_block_indx = [int(i) for i in block.split(".")]
                path = path[: -len(block)].rstrip("/")

        entry = await get_entry(
            path,
            ["read:data"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            metrics=request.state.metrics,
            structure_families={
                StructureFamily.array,
                StructureFamily.sparse,
                StructureFamily.table,
                StructureFamily.container,
            },
            access_policy=getattr(request.app.state, "access_policy", None),
        )

        if entry.structure_family == StructureFamily.container:
            # List the contents of a "simulated" zarr directory (excluding .zarray and .zgroup files)
            if hasattr(entry, "keys_range"):
                keys = await entry.keys_range(offset=0, limit=None)
            else:
                keys = entry.keys()
            url = str(request.url).rstrip("/")
            body = json.dumps([url + "/" + key for key in keys])

            return Response(body, status_code=200)

        elif entry.structure_family == StructureFamily.table:
            # List the columns of the table -- they will be accessed separately as arrays
            url = str(request.url).rstrip("/")
            body = json.dumps([url + "/" + key for key in entry.structure().columns])

            return Response(body, status_code=200)

        elif entry.structure_family in {StructureFamily.array, StructureFamily.sparse}:
            # Return the actual array values for a single block of zarr array
            if zarr_block_indx is not None:
                import numpy as np
                from sparse import SparseArray

                zarr_block_spec = convert_chunks_for_zarr(entry.structure().chunks)
                if (not (zarr_block_spec == [] and zarr_block_indx == [0])) and (
                    len(zarr_block_spec) != len(zarr_block_indx)
                ):
                    # Not a scalar and shape doesn't match
                    raise HTTPException(
                        status_code=HTTP_400_BAD_REQUEST,
                        detail=f"Requested zarr block index {zarr_block_indx} is inconsistent with the shape of array, {entry.structure().shape}.",  # noqa
                    )

                # Indices of the array slices in each dimension that correspond to the requested zarr block
                block_slices = tuple(
                    [
                        slice(i * c, (i + 1) * c)
                        for i, c in zip(zarr_block_indx, zarr_block_spec)
                    ]
                )
                try:
                    with record_timing(request.state.metrics, "read"):
                        array = await ensure_awaitable(entry.read, slice=block_slices)
                except IndexError:
                    raise HTTPException(
                        status_code=HTTP_400_BAD_REQUEST,
                        detail=f"Index of zarr block {zarr_block_indx} is out of range.",
                    )

                if isinstance(array, SparseArray):
                    array = array.todense()

                # Pad the last slices with zeros if needed to ensure all zarr blocks have same shapes
                padding_size = [
                    max(0, sl.stop - sh)
                    for sl, sh in zip(block_slices, entry.structure().shape)
                ]
                if sum(padding_size) > 0:
                    array = np.pad(
                        array, [(0, p) for p in padding_size], mode="constant"
                    )

                # Ensure the array is contiguous and encode it; equivalent to `buf=zarr.array(array).store['0.0']`
                array = array.astype(array.dtype, order=ZARR_BYTE_ORDER, copy=False)
                buf = zarr_codec.encode(array)
                if not isinstance(buf, bytes):
                    buf = array.tobytes(order="A")

                return Response(buf, status_code=200)

            else:
                # TODO:
                # Entire array (root uri) is requested -- never happens, but need to decide what to return here
                return Response(json.dumps({}), status_code=200)

    return router


def get_zarr_router_v3() -> APIRouter:
    router = APIRouter()

    @router.get("/{path:path}/zarr.json", name="Zarr v3 group or array metadata")
    @router.get("/{path:path}zarr.json", name="Zarr v3 group or array metadata")
    async def get_zarr_metadata(
        request: Request,
        path: str,
        principal: Union[Principal] = Depends(get_current_principal),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        root_tree: pydantic_settings.BaseSettings = Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
    ):
        from zarr.dtype import parse_data_type

        entry = await get_entry(
            path,
            ["read:data", "read:metadata"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            metrics=request.state.metrics,
            structure_families={
                StructureFamily.array,
                StructureFamily.table,
                StructureFamily.sparse,
                StructureFamily.container,
            },
            access_policy=getattr(request.app.state, "access_policy", None),
        )

        # Array or sparse array
        if entry.structure_family in {StructureFamily.array, StructureFamily.sparse}:
            structure = entry.structure()
            zarr_dtype = parse_data_type(
                structure.data_type.to_numpy_descr(), zarr_format=3
            )
            fill_value = (
                zarr_dtype.default_scalar() if structure.shape else entry.read()[()]
            )
            result = {
                "zarr_format": 3,
                "node_type": "array",
                "shape": list(structure.shape),
                "data_type": zarr_dtype.to_json(zarr_format=3),
                "chunk_grid": {
                    "name": "regular",
                    "configuration": {
                        "chunk_shape": convert_chunks_for_zarr(structure.chunks)
                    },
                },
                "chunk_key_encoding": {
                    "name": "default",
                    "configuration": {"separator": "/"},
                },
                "fill_value": zarr_dtype.to_json_scalar(fill_value, zarr_format=3),
                "codecs": [
                    {"name": "bytes", "configuration": {"endian": "little"}},
                    {
                        "name": zarr_codec.codec_id,
                        "configuration": {
                            "typesize": zarr_dtype.item_size,
                            **{
                                k: v
                                if k != "shuffle"
                                else {1: "shuffle", 0: "no_shuffle", 2: "bitshuffle"}[v]
                                for k, v in zarr_codec.get_config().items()
                                if k != "id"
                            },
                        },
                    },
                ],
                "dimension_names": list(structure.dims) if structure.dims else None,
                "attributes": entry.metadata(),
            }

        elif entry.structure_family in {
            StructureFamily.container,
            StructureFamily.table,
        }:
            # Structured numpy array, Container, or Table
            result = {
                "zarr_format": 3,
                "node_type": "group",
                "attributes": entry.metadata(),
            }

        return Response(
            orjson.dumps(result, option=orjson.OPT_SERIALIZE_NUMPY),
            status_code=200,
        )

    @router.get("/{path:path}/c/{block:path}", name="A chunk of a zarr array")
    async def get_zarr_array(
        request: Request,
        path: str,
        block: str,
        principal: Union[Principal] = Depends(get_current_principal),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        root_tree: pydantic_settings.BaseSettings = Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
    ):
        entry = await get_entry(
            path,
            ["read:data"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            metrics=request.state.metrics,
            structure_families={
                StructureFamily.array,
                StructureFamily.sparse,
            },
            access_policy=getattr(request.app.state, "access_policy", None),
        )

        if block is not None:
            import numpy as np
            from sparse import SparseArray

            zarr_block_indx = [int(i) for i in block.split("/")]
            zarr_block_spec = convert_chunks_for_zarr(entry.structure().chunks)
            if (not (zarr_block_spec == [] and zarr_block_indx == [0])) and (
                len(zarr_block_spec) != len(zarr_block_indx)
            ):
                # Not a scalar and shape doesn't match
                raise HTTPException(
                    status_code=HTTP_400_BAD_REQUEST,
                    detail=f"Requested zarr block index {zarr_block_indx} is inconsistent with the shape of array, {entry.structure().shape}.",  # noqa
                )

            # Indices of the array slices in each dimension that correspond to the requested zarr block
            block_slices = tuple(
                [
                    slice(i * c, (i + 1) * c)
                    for i, c in zip(zarr_block_indx, zarr_block_spec)
                ]
            )
            try:
                with record_timing(request.state.metrics, "read"):
                    array = await ensure_awaitable(entry.read, slice=block_slices)
            except IndexError:
                raise HTTPException(
                    status_code=HTTP_400_BAD_REQUEST,
                    detail=f"Index of zarr block {zarr_block_indx} is out of range.",
                )

            if isinstance(array, SparseArray):
                array = array.todense()

            # Pad the last slices with zeros if needed to ensure all zarr blocks have same shapes
            padding_size = [
                max(0, sl.stop - sh)
                for sl, sh in zip(block_slices, entry.structure().shape)
            ]
            if sum(padding_size) > 0:
                array = np.pad(array, [(0, p) for p in padding_size], mode="constant")

            # Ensure the array is contiguous and encode it; equivalent to `buf=zarr.array(array).store['0.0']`
            array = array.astype(array.dtype, order=ZARR_BYTE_ORDER, copy=False)
            buf = zarr_codec.encode(array)
            if not isinstance(buf, bytes):
                buf = array.tobytes(order="A")

            return Response(buf, status_code=200)

        else:
            # TODO:
            # Entire array (root uri) is requested -- never happens, but need to decide what to return here
            return Response(json.dumps({}), status_code=200)

    @router.get("/{path:path}", name="Contents of a zarr group")
    async def get_zarr_group(
        request: Request,
        path: str,
        principal: Union[Principal] = Depends(get_current_principal),
        authn_access_tags: Optional[Set[str]] = Depends(get_current_access_tags),
        authn_scopes: Scopes = Depends(get_current_scopes),
        root_tree: pydantic_settings.BaseSettings = Depends(get_root_tree),
        session_state: dict = Depends(get_session_state),
    ):
        entry = await get_entry(
            path,
            ["read:data"],
            principal,
            authn_access_tags,
            authn_scopes,
            root_tree,
            session_state,
            metrics=request.state.metrics,
            structure_families={
                StructureFamily.array,
                StructureFamily.container,
                StructureFamily.sparse,
                StructureFamily.table,
            },
            access_policy=getattr(request.app.state, "access_policy", None),
        )
        # Remove query params and the trailing slash from the url
        url = str(request.url).split("?")[0].rstrip("/")

        if entry.structure_family == StructureFamily.container:
            # List the contents of a "simulated" zarr directory (excluding .zarray and .zgroup files)
            if hasattr(entry, "keys_range"):
                keys = await entry.keys_range(offset=0, limit=None)
            else:
                keys = entry.keys()
            body = json.dumps([url + "/" + key for key in keys])

            return Response(body, status_code=200)

        elif entry.structure_family == StructureFamily.table:
            # List the columns of the table -- they will be accessed separately as arrays
            body = json.dumps([url + "/" + key for key in entry.structure().columns])

            return Response(body, status_code=200)

        else:
            return await get_zarr_metadata(
                request,
                path,
                principal=principal,
                authn_access_tags=authn_access_tags,
                authn_scopes=authn_scopes,
                root_tree=root_tree,
                session_state=session_state,
            )

    return router



import json
import logging
import os
import threading
import time
from urllib.parse import urlparse

import msgpack
import numpy as np
import websocket

from locust import HttpUser, between, events, task

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


@events.init_command_line_parser.add_listener
def _(parser):
    parser.add_argument(
        "--api-key",
        type=str,
        default="secret",
        help="API key for Tiled authentication (default: secret)",
    )
    parser.add_argument(
        "--node-name",
        type=str,
        required=True,
        help="Node name for streaming test (required)",
    )


@events.init.add_listener
def on_locust_init(environment, **kwargs):
    if environment.host is None:
        raise ValueError(
            "Host must be specified with --host argument, or through the web-ui."
        )

    # Create the streaming node once for all users
    create_streaming_node(
        environment.host,
        environment.parsed_options.api_key,
        environment.parsed_options.node_name,
    )


def create_streaming_node(host, api_key, node_name):
    """Create a streaming array node using Tiled client"""
    from tiled.client import from_uri

    # Connect to Tiled server using client
    client = from_uri(host, api_key=api_key)

    # Create initial streaming array
    arr = np.full(5, 0.0, dtype=np.float64)  # Initial array with zeros
    client.write_array(arr, key=node_name)

    logger.info(f"Created streaming node: {node_name}")
    client.logout()


class WriterUser(HttpUser):
    """User that writes streaming data to a Tiled node"""

    wait_time = between(0.1, 0.2)  # Wait 0.1-0.2 seconds between writes
    weight = int(os.getenv("WRITER_WEIGHT", 1))

    def on_start(self):
        """Initialize user state"""
        self.node_name = self.environment.parsed_options.node_name
        self.message_count = 0
        self.api_key = self.environment.parsed_options.api_key

        # Set authentication header
        self.client.headers.update({"Authorization": f"Apikey {self.api_key}"})

    @task(10)  # Run 10x as often as cleanup
    def write_data(self):
        """Write streaming data to the node"""
        # Create data with current timestamp as all values
        current_time = time.time()
        data = np.full(5, current_time, dtype=np.float64)
        binary_data = data.tobytes()

        # Post data to the streaming endpoint
        response = self.client.put(
            f"/api/v1/array/full/{self.node_name}",
            data=binary_data,
            headers={"Content-Type": "application/octet-stream"},
        )

        # Log status
        if response.status_code == 200:
            logger.debug(f"Wrote message {self.message_count} to node {self.node_name}")
            self.message_count += 1
        else:
            logger.error(
                f"Failed to write message {self.message_count}: {response.status_code} - {response.text}"
            )

    @task(1)
    def cleanup(self):
        """Periodically cleanup the stream"""
        if self.message_count > 50:
            # Close the stream
            response = self.client.delete(f"/api/v1/stream/close/{self.node_name}")
            if response.status_code == 200:
                logger.info(f"Closed stream for node {self.node_name}")

            # Reset message count (node persists for other users)
            self.message_count = 0


class StreamingUser(HttpUser):
    """User that connects to websocket stream and measures latency"""

    wait_time = between(1, 2)
    weight = int(os.getenv("STREAMING_WEIGHT", 1))

    def on_start(self):
        """Connect to the streaming endpoint"""
        self.node_name = self.environment.parsed_options.node_name
        self.api_key = self.environment.parsed_options.api_key
        self.envelope_format = "msgpack"  # Use msgpack for efficiency
        self.ws = None
        self.connected = False

        # Set up authentication for HTTP requests
        self.client.headers.update({"Authorization": f"Apikey {self.api_key}"})

        self._connect_websocket()

    def _connect_websocket(self):
        """Connect to the websocket stream"""
        try:
            # Parse host to get websocket URL
            parsed = urlparse(self.host)
            ws_scheme = "wss" if parsed.scheme == "https" else "ws"
            host = f"{ws_scheme}://{parsed.netloc}"

            ws_url = f"{host}/api/v1/stream/single/{self.node_name}?envelope_format={self.envelope_format}&start=0"

            # Create websocket connection
            self.ws = websocket.WebSocketApp(
                ws_url,
                header=[
                    f"Authorization: Apikey {self.api_key}"
                ],  # Proper Apikey format for websockets
                on_open=self._on_open,
                on_message=self._on_message,
                on_error=self._on_error,
                on_close=self._on_close,
            )

            # Start websocket in background thread
            self.ws_thread = threading.Thread(target=self.ws.run_forever)
            self.ws_thread.daemon = True
            self.ws_thread.start()

            # Wait a bit for connection
            time.sleep(0.5)

        except Exception as e:
            logger.error(f"Failed to connect websocket: {e}")

    def _on_open(self, ws):
        """Websocket connection opened"""
        self.connected = True
        logger.info(f"WebSocket connected to {self.node_name}")

    def _on_message(self, ws, message):
        """Process websocket messages and measure latency"""
        try:
            received_time = time.time()

            if isinstance(message, bytes):
                data = msgpack.unpackb(message)
            else:
                data = json.loads(message)

            # Extract timestamp from the payload (first element of the array)
            payload = data.get("payload")
            if payload and len(payload) > 0:
                # Convert bytes back to numpy array to get the timestamp
                payload_array = np.frombuffer(payload, dtype=np.float64)
                if len(payload_array) > 0:
                    write_time = payload_array[0]
                    latency_ms = (received_time - write_time) * 1000

                    logger.debug(
                        f"WS latency (sequence {data.get('sequence', 'N/A')}): {latency_ms:.1f}ms"
                    )

                    # Report to Locust
                    events.request.fire(
                        request_type="WS",
                        name="write_to_websocket_delivery",
                        response_time=latency_ms,
                        response_length=len(message),
                        exception=None,
                    )

        except Exception as e:
            logger.error(f"Error processing message: {e}")
            events.request.fire(
                request_type="WS",
                name="write_to_websocket_delivery",
                response_time=0,
                response_length=0,
                exception=e,
            )

    def _on_error(self, ws, error):
        """Websocket error occurred"""
        logger.error(f"WebSocket error: {error}")
        self.connected = False

    def _on_close(self, ws, close_status_code, close_msg):
        """Websocket connection closed"""
        logger.info(f"WebSocket closed: {close_status_code} - {close_msg}")
        self.connected = False

    @task
    def keep_alive(self):
        """Dummy task to keep the user active while listening for messages"""
        if not self.connected and self.ws:
            # Try to reconnect if disconnected
            logger.info("Attempting to reconnect WebSocket...")
            self._connect_websocket()

    def on_stop(self):
        """Clean up websocket connection"""
        if self.ws:
            self.ws.close()
            logger.info("WebSocket connection closed")



import copy
from pathlib import Path
from typing import Any, List, Optional, Union
from urllib.parse import quote_plus

import dask.dataframe
import pandas

from ..catalog.orm import Node
from ..storage import FileStorage, Storage
from ..structures.core import Spec, StructureFamily
from ..structures.data_source import Asset, DataSource
from ..structures.table import TableStructure
from ..type_aliases import JSON
from ..utils import path_from_uri
from .array import ArrayAdapter
from .dataframe import DataFrameAdapter
from .utils import init_adapter_from_catalog


class ParquetDatasetAdapter:
    """ """

    structure_family = StructureFamily.table
    supported_storage = {FileStorage}

    def __init__(
        self,
        data_uris: List[str],
        structure: TableStructure,
        metadata: Optional[JSON] = None,
        specs: Optional[List[Spec]] = None,
    ) -> None:
        """

        Parameters
        ----------
        data_uris :
        structure :
        metadata :
        specs :
        """
        # TODO Store data_uris instead and generalize to non-file schemes.
        if isinstance(data_uris, str):
            data_uris = [data_uris]
        self._partition_paths = [path_from_uri(uri) for uri in data_uris]
        self._metadata = metadata or {}
        self._structure = structure
        self.specs = list(specs or [])

    @classmethod
    def from_catalog(
        cls,
        data_source: DataSource[TableStructure],
        node: Node,
        /,
        **kwargs: Optional[Any],
    ) -> "ParquetDatasetAdapter":
        return init_adapter_from_catalog(cls, data_source, node, **kwargs)  # type: ignore

    def metadata(self) -> JSON:
        return self._metadata

    @property
    def dataframe_adapter(self) -> DataFrameAdapter:
        partitions = []
        for path in self._partition_paths:
            if not Path(path).exists():
                partition = None
            else:
                partition = dask.dataframe.read_parquet(path)
            partitions.append(partition)
        return DataFrameAdapter(
            partitions, self._structure, specs=self.specs, metadata=self.metadata()
        )

    @classmethod
    def init_storage(
        cls,
        storage: Storage,
        data_source: DataSource[TableStructure],
        path_parts: List[str],
    ) -> DataSource[TableStructure]:
        """

        Parameters
        ----------
        data_uri :
        structure :

        Returns
        -------

        """
        data_source = copy.deepcopy(data_source)  # Do not mutate caller input.
        data_uri = storage.uri + "".join(
            f"/{quote_plus(segment)}" for segment in path_parts
        )
        directory = path_from_uri(data_uri)
        directory.mkdir(parents=True, exist_ok=True)
        assets = [
            Asset(
                data_uri=f"{data_uri}/partition-{i}.parquet",
                is_directory=False,
                parameter="data_uris",
                num=i,
            )
            for i in range(data_source.structure.npartitions)
        ]
        data_source.assets.extend(assets)
        return data_source

    def write_partition(
        self, data: Union[dask.dataframe.DataFrame, pandas.DataFrame], partition: int
    ) -> None:
        """

        Parameters
        ----------
        data :
        partition :

        Returns
        -------

        """
        uri = self._partition_paths[partition]
        data.to_parquet(uri)

    def write(self, data: Union[dask.dataframe.DataFrame, pandas.DataFrame]) -> None:
        """

        Parameters
        ----------
        data :

        Returns
        -------

        """
        if self.structure().npartitions != 1:
            raise NotImplementedError
        uri = self._partition_paths[0]
        data.to_parquet(uri)

    def read(self, *args: Any, **kwargs: Any) -> pandas.DataFrame:
        """

        Parameters
        ----------
        args :
        kwargs :

        Returns
        -------

        """
        return self.dataframe_adapter.read(*args, **kwargs)

    def read_partition(self, *args: Any, **kwargs: Any) -> pandas.DataFrame:
        """

        Parameters
        ----------
        args :
        kwargs :

        Returns
        -------

        """
        return self.dataframe_adapter.read_partition(*args, **kwargs)

    def structure(self) -> TableStructure:
        """

        Returns
        -------

        """
        return self._structure

    def get(self, key: str) -> Union[ArrayAdapter, None]:
        return self.dataframe_adapter.get(key)



import warnings
from collections import defaultdict
from typing import TYPE_CHECKING, Any, Optional

from ..structures.data_source import DataSource

if TYPE_CHECKING:
    from ..type_aliases import AnyStructure

# for back-compat
from ..utils import node_repr as tree_repr  # noqa: F401

_MESSAGE = (
    "Instead of {name}_indexer[...] use {name}()[...]. "
    "The {name}_indexer accessor is deprecated."
)


class IndexersMixin:
    """
    Provides sliceable attributes keys_indexer, items_indexer, values_indexer.

    This is just for back-ward compatibility.
    """

    keys: Any
    values: Any
    items: Any
    fn: Any

    @property
    def keys_indexer(self) -> Any:
        warnings.warn(_MESSAGE.format(name="keys"), DeprecationWarning)
        return self.keys()

    @property
    def values_indexer(self) -> Any:
        warnings.warn(_MESSAGE.format(name="values"), DeprecationWarning)
        return self.values()

    @property
    def items_indexer(self) -> Any:
        warnings.warn(_MESSAGE.format(name="items"), DeprecationWarning)
        return self.items()


class IndexCallable:
    """
    DEPRECATED and no longer used internally

    Provide getitem syntax for functions

    >>> def inc(x):
    ...     return x + 1

    >>> I = IndexCallable(inc)
    >>> I[3]
    4

    Vendored from dask
    """

    __slots__ = ("fn",)

    def __init__(self, fn: Any) -> None:
        self.fn = fn

    def __getitem__(self, key: str) -> Any:
        return self.fn(key)


def asset_parameters_to_adapter_kwargs(
    data_source: DataSource["AnyStructure"],
) -> dict[str, Any]:
    """Transform database representation of Adapter parameters to Python representation."""
    parameters: dict[str, Any] = defaultdict(list)
    for asset in data_source.assets:
        if (asset.num is not None) or (asset.parameter == "data_uris"):
            # This asset is associated with a parameter that takes a list of URIs.
            param = asset.parameter or "data_uris"
            parameters[param].append(asset.data_uri)
        else:
            # This asset is associated with a parameter that takes a single URI.
            param = asset.parameter or "data_uri"
            parameters[param] = asset.data_uri

    return parameters


def init_adapter_from_catalog(
    adapter_cls: type[Any],
    data_source: DataSource["AnyStructure"],
    node: Any,  # tiled.catalog.orm.Node ?
    /,
    **kwargs: Optional[Any],
) -> Any:
    # TODO: Sort out typing for Adapters
    """Factory function to produce Adapter instances given their parameters encoded in data sources"""
    parameters = asset_parameters_to_adapter_kwargs(data_source)
    kwargs.update(parameters)
    kwargs["metadata"] = node.metadata_
    kwargs["specs"] = node.specs
    return adapter_cls(structure=data_source.structure, **kwargs)



"""
This module was created to save structure methods that can be handled by
the server and the client.

"""

import enum
import importlib
from dataclasses import asdict, dataclass
from typing import Dict, Optional

from pydantic import StringConstraints
from typing_extensions import Annotated

from ..utils import OneShotCachedMap


class StructureFamily(str, enum.Enum):
    array = "array"
    awkward = "awkward"
    container = "container"
    sparse = "sparse"
    table = "table"


@dataclass(frozen=True)
class Spec:
    name: Annotated[str, StringConstraints(max_length=255)]
    version: Optional[Annotated[str, StringConstraints(max_length=255)]] = None

    def __init__(self, name, version=None) -> None:
        # Enable the name to be passed as a position argument.
        # The setattr stuff is necessary to make this work with a frozen dataclass.
        object.__setattr__(self, "name", name)
        object.__setattr__(self, "version", version)

    def __repr__(self):
        # Display the name as a positional argument, for conciseness.
        if self.version is None:
            output = f"{type(self).__name__}({self.name!r})"
        else:
            output = f"{type(self).__name__}({self.name!r}, version={self.version!r})"
        return output

    def dict(self) -> Dict[str, Optional[str]]:
        # For easy interoperability with pydantic 1.x models
        return asdict(self)

    model_dump = dict  # For easy interoperability with pydantic 2.x models


# TODO: make type[Structure] after #1036
STRUCTURE_TYPES = OneShotCachedMap[StructureFamily, type](
    {
        StructureFamily.array: lambda: importlib.import_module(
            "...structures.array", StructureFamily.__module__
        ).ArrayStructure,
        StructureFamily.awkward: lambda: importlib.import_module(
            "...structures.awkward", StructureFamily.__module__
        ).AwkwardStructure,
        StructureFamily.table: lambda: importlib.import_module(
            "...structures.table", StructureFamily.__module__
        ).TableStructure,
        StructureFamily.sparse: lambda: importlib.import_module(
            "...structures.sparse", StructureFamily.__module__
        ).SparseStructure,
    }
)



import contextlib
import copy
import pathlib
import platform
import secrets
import shutil
import tempfile
import threading
import time
from typing import Optional, Union, cast
from urllib.parse import quote_plus, urlparse

import uvicorn

from ..storage import SQLStorage, get_storage
from ..utils import ensure_uri

_server_is_running = False


class ThreadedServer(uvicorn.Server):
    def install_signal_handlers(self):
        pass

    @contextlib.contextmanager
    def run_in_thread(self):
        global _server_is_running

        if _server_is_running:
            raise RuntimeError(
                "Only one server can be run at a time " "in a given Python process."
            )
        _server_is_running = True
        thread = threading.Thread(target=self.run)
        thread.start()
        try:
            # Wait for server to start up, or raise TimeoutError.
            for _ in range(200):
                time.sleep(0.1)
                if self.started:
                    break
            else:
                raise TimeoutError("Server did not start in 20 seconds.")
            host, port = self.servers[0].sockets[0].getsockname()
            yield f"http://{host}:{port}"
        finally:
            self.should_exit = True
            thread.join()
            _server_is_running = False


class SimpleTiledServer:
    """
    Run a simple Tiled server on a background thread.

    This is intended to be used for tutorials and development. It employs only
    basic security and should not be used to store anything important. It does
    not scale to large number of users. By default, it uses temporary storage.

    Parameters
    ----------
    directory : Optional[Path, str]
        Location where data and embedded databases will be stored.
        By default, a temporary directory will be used.
    api_key : Optional[str]
        By default, an 8-bit random secret is generated. (Production Tiled
        servers use longer secrets.)
    port : Optional[int]
        Port the server will listen on. By default, a random free high port
        is allocated by the operating system.

    Examples
    --------

    Run a server and connect to it.

    >>> from tiled.server import SimpleTiledServer
    >>> from tiled.client import from_uri
    >>> ts = SimpleTiledServer()
    >>> client = from_uri(ts.uri)

    Locate server data, databases, and log files.

    >>> ts.directory

    Run a server with persistent storage that can be reused.

    >>> ts = SimpleTiledServer("my_data/")
    """

    def __init__(
        self,
        directory: Optional[Union[str, pathlib.Path]] = None,
        api_key: Optional[str] = None,
        port: int = 0,
        readable_storage: Optional[Union[str, pathlib.Path]] = None,
    ):
        # Delay import to avoid circular import.
        from ..catalog import from_uri as catalog_from_uri
        from .app import build_app
        from .logging_config import LOGGING_CONFIG

        if directory is None:
            directory = pathlib.Path(tempfile.mkdtemp())
            self._cleanup_directory = True
        else:
            directory = pathlib.Path(directory).resolve()
            self._cleanup_directory = False
        (directory / "data").mkdir(parents=True, exist_ok=True)
        storage_uri = ensure_uri(f"duckdb:///{str(directory / 'storage.duckdb')}")

        # In production we use a proper 32-bit token, but for brevity we
        # use just 8 here. This server only accepts connections on localhost
        # and is not intended for production use, so we think this is an
        # acceptable concession to usability.
        api_key = api_key or secrets.token_hex(8)

        # Alter copy of default LOGGING_CONFIG to log to files instead of
        # stdout and stderr.
        log_config = copy.deepcopy(LOGGING_CONFIG)
        log_config["handlers"]["access"]["class"] = "logging.FileHandler"
        del log_config["handlers"]["access"]["stream"]
        log_config["handlers"]["access"]["filename"] = str(directory / "access.log")
        log_config["handlers"]["default"]["class"] = "logging.FileHandler"
        del log_config["handlers"]["default"]["stream"]
        log_config["handlers"]["default"]["filename"] = str(directory / "error.log")

        self.catalog = catalog_from_uri(
            directory / "catalog.db",
            writable_storage=[directory / "data", storage_uri],
            init_if_not_exists=True,
            readable_storage=readable_storage,
        )
        self.app = build_app(
            self.catalog, authentication={"single_user_api_key": api_key}
        )
        self._cm = ThreadedServer(
            uvicorn.Config(self.app, port=port, loop="asyncio", log_config=log_config)
        ).run_in_thread()
        base_url = self._cm.__enter__()

        # Extract port from base_url.
        actual_port = urlparse(base_url).port

        # Stash attributes for easy introspection
        self.port = actual_port
        self.directory = directory
        self.storage = cast(SQLStorage, get_storage(storage_uri))
        self.api_key = api_key
        self.uri = f"{base_url}/api/v1?api_key={quote_plus(api_key)}"
        self.web_ui_link = f"{base_url}?api_key={quote_plus(api_key)}"

    def __repr__(self):
        return f"<{type(self).__name__} '{self.uri}'>"

    def _repr_html_(self):
        # For Jupyter
        return f"""
<table>
  <tr>
    <td>Web Interface</td>
    <td><a href={self.web_ui_link}>{self.web_ui_link}</a></td>
  </tr>
    <td>API</td>
    <td><code>{self.web_ui_link}</code></td>
  </tr>
</table>"""

    def close(self):
        self.storage.dispose()  # Close the connection to the storage DB
        self._cm.__exit__(None, None, None)
        if self._cleanup_directory and (platform.system() != "Windows"):
            # Windows cannot delete the logfiles because the global Python
            # logging system still has the logfiles open for appending.
            shutil.rmtree(self.directory)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()



import subprocess
import sys

import httpx
import numpy
import pytest
import uvicorn
from fastapi import APIRouter
from starlette.status import HTTP_500_INTERNAL_SERVER_ERROR

from ..adapters.array import ArrayAdapter
from ..adapters.mapping import MapAdapter
from ..catalog import in_memory
from ..client import from_uri
from ..server.app import build_app, build_app_from_config
from ..server.logging_config import LOGGING_CONFIG
from .utils import Server

router = APIRouter()

API_KEY = "secret"


@pytest.fixture
def server(tmpdir):
    catalog = in_memory(writable_storage=str(tmpdir))
    app = build_app(catalog, {"single_user_api_key": API_KEY})
    app.include_router(router)
    config = uvicorn.Config(app, port=0, loop="asyncio", log_config=LOGGING_CONFIG)
    server = Server(config)
    with server.run_in_thread() as url:
        yield url


@pytest.fixture
def public_server(tmpdir):
    catalog = in_memory(writable_storage=str(tmpdir))
    app = build_app(
        catalog, {"single_user_api_key": API_KEY, "allow_anonymous_access": True}
    )
    app.include_router(router)
    config = uvicorn.Config(app, port=0, loop="asyncio", log_config=LOGGING_CONFIG)
    server = Server(config)
    with server.run_in_thread() as url:
        yield url


arr = ArrayAdapter.from_array(numpy.ones((5, 5)))
tree = MapAdapter({"A1": arr, "A2": arr})


@pytest.fixture
def multiuser_server(tmpdir):
    database_uri = f"sqlite:///{tmpdir}/tiled.sqlite"
    subprocess.run(
        [sys.executable, "-m", "tiled", "admin", "initialize-database", database_uri],
        check=True,
        capture_output=True,
    )
    config = {
        "authentication": {
            "secret_keys": ["SECRET"],
            "providers": [
                {
                    "provider": "toy",
                    "authenticator": "tiled.authenticators:DictionaryAuthenticator",
                    "args": {
                        "users_to_passwords": {"alice": "secret1", "bob": "secret2"}
                    },
                }
            ],
        },
        "database": {
            "uri": database_uri,
        },
        "trees": [
            {
                "tree": f"{__name__}:tree",
                "path": "/",
            },
        ],
    }
    app = build_app_from_config(config)
    app.include_router(router)
    config = uvicorn.Config(app, port=0, loop="asyncio", log_config=LOGGING_CONFIG)
    server = Server(config)
    with server.run_in_thread() as url:
        yield url


@router.get("/error")
def error():
    1 / 0  # type: ignore error!


@pytest.mark.filterwarnings("ignore: websockets.legacy is deprecated")
@pytest.mark.filterwarnings(
    "ignore: websockets.server.WebSocketServerProtocol is deprecated"
)
def test_500_response(server):
    """
    Test that unexpected server error returns 500 response.

    This test is meant to catch regressions in which server exceptions can
    result in the server sending no response at all, leading clients to raise
    like:

    httpx.RemoteProtocolError: Server disconnected without sending a response.

    This can happen when bugs are introduced in the middleware layer.
    """
    client = from_uri(server, api_key=API_KEY)
    response = client.context.http_client.get(f"{server}/error")
    assert response.status_code == HTTP_500_INTERNAL_SERVER_ERROR


def test_writing_integration(server):
    client = from_uri(server, api_key=API_KEY)
    x = client.write_array([1, 2, 3], key="array")
    x[:]


def test_public_server(public_server):
    from_uri(public_server)


def test_internal_authentication_mode_with_password_clients(multiuser_server):
    "The 'internal' authentication mode used to be named 'password'."
    # Mock old client
    response = httpx.get(
        multiuser_server + "/api/v1/", headers={"user-agent": "python-tiled/0.1.0b16"}
    )
    actual_mode = response.json()["authentication"]["providers"][0]["mode"]
    assert actual_mode == "password"

    # Mock new client
    response = httpx.get(
        multiuser_server + "/api/v1/", headers={"user-agent": "python-tiled/0.1.0b17"}
    )
    actual_mode = response.json()["authentication"]["providers"][0]["mode"]
    assert actual_mode == "internal"

    # Mock unknown client
    response = httpx.get(multiuser_server + "/api/v1/", headers={})
    actual_mode = response.json()["authentication"]["providers"][0]["mode"]
    assert actual_mode == "internal"
